from typing import TYPE_CHECKING, Any
from .cloneable import Cloneable

if TYPE_CHECKING:
    from boxsdk.session.session import Session
    from boxsdk.util.translator import Translator


class BaseEndpoint(Cloneable):
    """A Box API endpoint."""

    def __init__(self, session: 'Session', **kwargs: Any):
        """
        :param session:
            The Box session used to make requests.
        :param kwargs:
            Keyword arguments for base class constructors.
        """
        super().__init__(**kwargs)
        self._session = session

    @property
    def session(self) -> 'Session':
        """
        Get the :class:`BoxSession` instance the object is using.
        """
        return self._session

    @property
    def translator(self) -> 'Translator':
        """
        The translator used for translating Box API JSON responses into `BaseAPIJSONObject` smart objects.
        """
        return self._session.translator

    def get_url(self, *args: Any) -> str:
        """
        Return the URL used to access the endpoint.

        :param args:
            Parts of the endpoint URL.
        """
        return self._session.get_url(*args)

    def clone(self, session: 'Session' = None) -> 'BaseEndpoint':
        """
        Returns a copy of this cloneable object using the specified session.

        :param session:
            The Box session used to make requests.
        """
        return self.__class__(session or self._session)



from typing import TYPE_CHECKING, List, Optional, Iterable, Any
from .base_object import BaseObject
from ..util.api_call_decorator import api_call
from ..util.text_enum import TextEnum

if TYPE_CHECKING:
    from boxsdk.session.session import Session


class MetadataTemplateUpdate:
    """Represents a set of update operations to a metadata template."""

    def __init__(self):
        super().__init__()
        self._ops = []

    def json(self) -> list:
        return self._ops

    def add_enum_option(self, field_key: str, option_key: str) -> None:
        """
        Adds a new option to an enum field.

        :param field_key:
            The key of the template field to add the option to
        :param option_key:
            The option to add
        """
        self.add_operation({
            'op': 'addEnumOption',
            'fieldKey': field_key,
            'data': {
                'key': option_key,
            },
        })

    def add_field(self, field: 'MetadataField') -> None:
        """
        Add a new field to the template.

        :param field:
            The new field to add
        """
        self.add_operation({
            'op': 'addField',
            'data': field.json(),
        })

    def edit_template(self, data: dict) -> None:
        """
        Edit top-level template properties.

        :param data:
            The properties to modify
        """
        self.add_operation({
            'op': 'editTemplate',
            'data': data,
        })

    def reorder_enum_options(self, field_key: str, option_keys: List[str]) -> None:
        """
        Reorders the options in an enum field, which affects their display in UI.

        :param field_key:
            The key of the enum field to reorder
        :param option_keys:
            The option keys in the desired order
        """
        self.add_operation({
            'op': 'reorderEnumOptions',
            'fieldKey': field_key,
            'enumOptionKeys': option_keys,
        })

    def reorder_fields(self, field_keys: List[str]) -> None:
        """
        Reorders the fields in a metadata template, which affects their display in UI.

        :param field_keys:
            The field keys in the desired order
        """
        self.add_operation({
            'op': 'reorderFields',
            'fieldKeys': field_keys,
        })

    def edit_field(self, field_key: str, field: 'MetadataField') -> None:
        """
        Edits a field in the template.

        :param field_key:
            The key of the field to update
        :param field:
            The updated field values
        """
        self.add_operation({
            'op': 'editField',
            'fieldKey': field_key,
            'data': field.json(),
        })

    def edit_enum_option_key(self, field_key: str, old_option_key: str, new_option_key: str) -> None:
        """
        Change the key of an enum field option.

        :param field_key:
            The key of the template field in which the option appears
        :param old_option_key:
            The old option key
        :param new_option_key:
            The new option key
        """
        self.add_operation({
            'op': 'editEnumOption',
            'fieldKey': field_key,
            'enumOptionKey': old_option_key,
            'data': {
                'key': new_option_key,
            },
        })

    def remove_enum_option(self, field_key: str, option_key: str) -> None:
        """
        Remove an option from an enum field.

        :param field_key:
            The key of the template field in which the option appears
        :param option_key:
            The key of the enum option to remove
        """
        self.add_operation({
            'op': 'removeEnumOption',
            'fieldKey': field_key,
            'enumOptionKey': option_key,
        })

    def remove_field(self, field_key: str) -> None:
        """
        Remove a field from the metadata template.

        :param field_key:
            The key of the field to remove
        """
        self.add_operation({
            'op': 'removeField',
            'fieldKey': field_key,
        })

    def add_operation(self, operation: dict) -> None:
        """
        Adds an update operation.

        :param operation:
            The operation to add.
        """
        self._ops.append(operation)


class MetadataFieldType(TextEnum):
    STRING = 'string'
    DATE = 'date'
    ENUM = 'enum'
    MULTISELECT = 'multiSelect'
    FLOAT = 'float'


class MetadataField:
    """Represents a metadata field when creating or updating a metadata template."""

    def __init__(
            self,
            field_type: MetadataFieldType,
            display_name: str,
            key: Optional[str] = None,
            options: Iterable[str] = None,
            description: Optional[str] = None,
            hidden: Optional[bool] = None
    ):
        """
        :param field_type:
            The type of the metadata field
        :param display_name:
            The human-readable name of the metadata field
        :param key:
            The machine-readable key for the metadata field
        :param options:
            For 'enum' or 'multiSelect' fields, the selectable options
        :param description:
            A description of the field
        :param hidden:
            Whether this field is hidden in the UI for the user and can only be set through the API instead
        """
        super().__init__()
        self.type = field_type
        self.name = display_name
        self.key = key
        self.options = options
        self.description = description
        self.hidden = hidden

    def json(self) -> dict:
        """
        Returns the correct representation of the template field for the API.
        """
        values = {}

        if self.type is not None:
            values['type'] = self.type

        if self.name is not None:
            values['displayName'] = self.name

        if self.key is not None:
            values['key'] = self.key

        if self.type in ['enum', 'multiSelect']:
            values['options'] = [{'key': opt} for opt in self.options or ()]

        if self.description is not None:
            values['description'] = self.description

        if self.hidden is not None:
            values['hidden'] = self.hidden

        return values


class MetadataTemplate(BaseObject):
    """Represents a metadata template, which contains the the type information for associated metadata fields."""

    _item_type = 'metadata_template'
    _untranslated_fields = ('fields',)
    _scope = None
    _template_key = None

    def __init__(self, session: 'Session', object_id: Optional[str], response_object: Optional[dict] = None):
        """
        :param session:
            The Box session used to make requests.
        :param object_id:
            The primary GUID key for the metadata template
        :param response_object:
            A JSON object representing the object returned from a Box API request.  This should
            contain 'scope' and 'templateKey' properties if the instance is being constructed without
            a primary GUID object_id.
        """
        super().__init__(session, object_id, response_object)
        if response_object:
            self._scope = response_object.get('scope', None)
            self._template_key = response_object.get('templateKey', None)
        elif not object_id:
            raise ValueError('Metadata template must be constructed with an ID or scope and templateKey')

    @property
    def scope(self) -> Optional[str]:
        return self._scope

    @property
    def template_key(self) -> Optional[str]:
        return self._template_key

    def get_url(self, *args: Any) -> str:
        """
        Base class override, since metadata templates have a weird compound ID and non-standard URL format
        """
        if self._scope and self._template_key:
            return self._session.get_url('metadata_templates', self._scope, self._template_key, 'schema', *args)

        return super().get_url(*args)

    @staticmethod
    def start_update() -> MetadataTemplateUpdate:
        """
        Start an update operation on the template.

        :returns:
            An update object to collect the desired update operations.
        """
        return MetadataTemplateUpdate()

    @api_call
    def update_info(self, *, updates: MetadataTemplateUpdate, **kwargs) -> 'MetadataTemplate':
        # pylint: disable=arguments-differ
        """
        Update a metadata template with a set of update operations.

        :param updates:
            The update operations to apply to the template
        :returns:
            The updated metadata template object
        """
        # pylint: disable=arguments-differ
        return super().update_info(data=updates.json(), **kwargs)



from unittest.mock import Mock
import pytest

from boxsdk.config import API
from boxsdk.object.device_pinner import DevicePinner
from boxsdk.network.default_network import DefaultNetworkResponse


@pytest.fixture(scope='module')
def delete_device_pin_response():
    # pylint:disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.ok = True
    return mock_network_response


def test_get(test_device_pin, mock_box_session):
    created_at = '2016-05-18T17:38:03-07:00'
    expected_url = f'{API.BASE_API_URL}/device_pinners/{test_device_pin.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'device_pinner',
        'id': test_device_pin.object_id,
        'created_at': created_at
    }
    device_pin = test_device_pin.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(device_pin, DevicePinner)
    assert device_pin.created_at == created_at


def test_delete_device_pin_return_the_correct_response(
        test_device_pin,
        mock_box_session,
        delete_device_pin_response,
):
    # pylint:disable=redefined-outer-name
    mock_box_session.delete.return_value = delete_device_pin_response
    response = test_device_pin.delete()
    # pylint:disable=protected-access
    expected_url = test_device_pin.get_url()
    # pylint:enable = protected-access
    mock_box_session.delete.assert_called_once_with(expected_url, params={}, expect_json_response=False, headers=None)
    assert response is True



from functools import partial
import json
from logging import Logger
from operator import attrgetter
from pprint import pformat
from unittest.mock import DEFAULT, Mock, patch, ANY

import pytest
from requests import Response

from boxsdk.network import default_network
from boxsdk.network.default_network import DefaultNetworkResponse, DefaultNetwork


class ExceptionSubclass(Exception):
    pass


@pytest.fixture
def logger():
    return Mock(Logger)


@pytest.fixture
def network():
    return DefaultNetwork()


@pytest.fixture(autouse=True)
def mocked_logger(logger):
    with patch.object(default_network, 'getLogger') as get_logger:
        get_logger.return_value = logger
        yield


@pytest.fixture(scope='module')
def logger_method_names():
    return ['critical', 'debug', 'error', 'exception', 'fatal', 'info', 'log', 'warn', 'warning']


@pytest.fixture
def logger_methods(logger, logger_method_names):
    return list(map(partial(getattr, logger), logger_method_names))


@pytest.fixture
def assert_logger_called_once_with(logger, logger_method_names):

    def _assert_logger_called_once_with(method_name, *args, **kwargs):
        getattr(logger, method_name).assert_called_once_with(*args, **kwargs)
        method_names = set(logger_method_names)
        method_names.discard(method_name)
        for name in method_names:
            getattr(logger, name).assert_not_called()

    return _assert_logger_called_once_with


@pytest.fixture
def assert_logger_not_called(logger_methods):

    def _assert_logger_not_called():
        for method in logger_methods:
            method.assert_not_called()

    return _assert_logger_not_called


@pytest.fixture
def logger_call_count(logger_methods):

    def _logger_call_count():
        return sum(map(attrgetter('call_count'), logger_methods))

    return _logger_call_count


@pytest.fixture
def make_network_request_and_assert_response(make_network_request, request_response):
    # pylint:disable=unused-argument

    def _make_network_request(*args, **kwargs):
        response = make_network_request(*args, **kwargs)
        assert response.request_response is request_response
        return response

    return _make_network_request


@pytest.fixture
def construct_network_response(access_token):

    def _construct_network_response(request_response, log_response_content=True):
        return DefaultNetworkResponse(
            request_response=request_response,
            access_token_used=access_token,
            log_response_content=log_response_content
        )

    return _construct_network_response


@pytest.fixture
def network_response(construct_network_response, request_response):
    return construct_network_response(request_response)


@pytest.fixture(params=['content', 'json'])
def get_content_from_response(request):

    def _get_content_from_response(response):
        content = getattr(response, request.param)
        if request.param in ['json']:
            try:
                content = content()
            except ValueError:
                pass

    return _get_content_from_response


@pytest.fixture(params=['response_as_stream', 'request_response', 'log'])
def do_not_get_content_from_response(request):

    def _do_not_get_content_from_response(response):
        attribute = getattr(response, request.param)
        if request.param in ['log']:
            attribute()

    return _do_not_get_content_from_response


# BEGIN Tests for DefaultNetwork.


def test_default_network_response_properties_pass_through_to_session_response_properties(access_token):
    mock_session_response = Mock(Response)
    mock_session_response.status_code = 200
    mock_session_response.headers = {}
    mock_session_response.raw = Mock()
    mock_session_response.content = json.dumps('content')
    mock_session_response.request = Mock()
    network_reponse = DefaultNetworkResponse(mock_session_response, access_token)
    assert network_reponse.json() == mock_session_response.json()
    assert network_reponse.content == mock_session_response.content
    assert network_reponse.ok == mock_session_response.ok
    assert network_reponse.status_code == mock_session_response.status_code
    assert network_reponse.headers == mock_session_response.headers
    assert network_reponse.response_as_stream == mock_session_response.raw
    assert network_reponse.access_token_used == access_token


def test_default_network_request(make_network_request_and_assert_response):
    # pylint:disable=redefined-outer-name
    default_network = DefaultNetwork()
    make_network_request_and_assert_response(default_network, custom_kwargs='test')


@pytest.mark.parametrize('delay', (0, 1))
def test_default_network_retry_after_sleeps(delay):
    default_network = DefaultNetwork()
    retry_call = Mock()
    mock_sleep = Mock()
    with patch('boxsdk.network.default_network.time.sleep', mock_sleep):
        default_network.retry_after(delay, retry_call, DEFAULT, kwarg=DEFAULT)
    mock_sleep.assert_called_once_with(delay)
    retry_call.assert_called_once_with(DEFAULT, kwarg=DEFAULT)


def test_network_response_constructor(make_network_request_and_assert_response):
    assert DefaultNetwork().network_response_constructor is DefaultNetworkResponse

    class DefaultNetworkResponseSubclass(DefaultNetworkResponse):
        pass

    class DefaultNetworkSubclass(DefaultNetwork):
        @property
        def network_response_constructor(self):
            return DefaultNetworkResponseSubclass

    network = DefaultNetworkSubclass()
    response = make_network_request_and_assert_response(network)
    assert isinstance(response, DefaultNetworkResponseSubclass)


def test_network_logs_requests(make_network_request, http_verb, test_url, network, logger):
    kwargs = dict(custom_kwarg='foo')
    make_network_request(network, **kwargs)
    logger.info.assert_any_call(
        network.REQUEST_FORMAT,
        {'method': http_verb, 'url': test_url, 'request_kwargs': pformat(kwargs)},
    )


def test_network_logs_request_exception(make_network_request, mock_request, http_verb, test_url, logger, network):
    mock_request.side_effect = expected_exception = ExceptionSubclass('exception raised from request()')
    with pytest.raises(ExceptionSubclass) as pytest_exc_info:
        make_network_request(network)
    assert pytest_exc_info.value is expected_exception
    logger.warning.assert_called_once_with(
        network.EXCEPTION_FORMAT,
        {'method': http_verb, 'url': test_url, 'exc_type_name': ExceptionSubclass.__name__, 'exc_value': expected_exception},
    )


def test_network_request_returns_network_response(make_network_request, request_response, network):
    response = make_network_request(network)
    assert response.request_response is request_response
    assert isinstance(response, DefaultNetworkResponse)


# BEGIN Tests for NetworkResponse.


def test_all_network_responses_log_immediately(network_response, logger_call_count):
    # pylint:disable=unused-argument
    # Need to load the `network_response` fixture for this test to be
    # meaningful.
    assert logger_call_count() == 1


def test_network_response_only_logs_once(network_response, logger_call_count):
    network_response.log()
    assert logger_call_count() == 1
    network_response.log()
    assert logger_call_count() == 1


@pytest.mark.parametrize('content_length_header', [False, True])
def test_network_logs_successful_json_responses(
        construct_network_response, generic_successful_request_response, assert_logger_called_once_with,
        get_content_from_response, http_verb, test_url, content_length_header,
):
    expected_content_length = str(len(generic_successful_request_response.content))
    if content_length_header:
        generic_successful_request_response.headers['Content-Length'] = expected_content_length
    else:
        generic_successful_request_response.headers.pop('Content-Length', None)
    generic_successful_request_response.request.method = http_verb
    generic_successful_request_response.request.url = test_url
    network_response = construct_network_response(generic_successful_request_response)
    get_content_from_response(network_response)
    assert_logger_called_once_with(
        'info',
        DefaultNetworkResponse.SUCCESSFUL_RESPONSE_FORMAT,
        {
            'method': http_verb,
            'url': test_url,
            'status_code': generic_successful_request_response.status_code,
            'content_length': expected_content_length,
            'headers': pformat(generic_successful_request_response.headers),
            'content': pformat(generic_successful_request_response.json()),
        }
    )


@pytest.mark.parametrize('content_length_header', [False, True])
def test_network_logs_successful_responses_with_file_content_placeholder(
        construct_network_response, generic_successful_request_response, assert_logger_called_once_with,
        logger, content_length_header,
):
    if content_length_header:
        expected_content_length = str(len(generic_successful_request_response.content))
        generic_successful_request_response.headers['Content-Length'] = expected_content_length
    else:
        generic_successful_request_response.headers.pop('Content-Length', None)
        expected_content_length = '?'
    network_response = construct_network_response(generic_successful_request_response, log_response_content=False)
    assert_logger_called_once_with('info', DefaultNetworkResponse.SUCCESSFUL_RESPONSE_FORMAT, ANY)
    assert logger.info.call_args[0][1]['content'] == network_response.CONTENT_NOT_LOGGED
    assert logger.info.call_args[0][1]['content_length'] == expected_content_length


def test_network_logs_successful_responses_with_non_json_content(
        construct_network_response, generic_successful_request_response, assert_logger_called_once_with,
        logger, get_content_from_response,
):
    generic_successful_request_response.content = content = (b''.join(chr(i).encode('utf-8') for i in range(128)) * 4)
    generic_successful_request_response.json.side_effect = lambda: json.loads(content.decode('utf-8'))
    network_response = construct_network_response(generic_successful_request_response, log_response_content=False)
    get_content_from_response(network_response)
    assert_logger_called_once_with('info', DefaultNetworkResponse.SUCCESSFUL_RESPONSE_FORMAT, ANY)
    assert logger.info.call_args[0][1]['content'] == network_response.CONTENT_NOT_LOGGED


def test_network_logs_non_successful_responses(
        construct_network_response, server_error_request_response, assert_logger_called_once_with,
        http_verb, test_url,
):
    server_error_request_response.request.method = http_verb
    server_error_request_response.request.url = test_url
    construct_network_response(server_error_request_response)
    assert_logger_called_once_with(
        'warning',
        DefaultNetworkResponse.ERROR_RESPONSE_FORMAT,
        {
            'method': http_verb,
            'url': test_url,
            'status_code': server_error_request_response.status_code,
            'content_length': str(len(server_error_request_response.content)),
            'headers': pformat(server_error_request_response.headers),
            'content': pformat(server_error_request_response.json()),
        }
    )


# END Tests for NetworkResponse.



import json
from typing import TYPE_CHECKING, Optional, Iterable

from boxsdk.object.base_object import BaseObject
from ..pagination.marker_based_object_collection import MarkerBasedObjectCollection
from ..util.api_call_decorator import api_call

if TYPE_CHECKING:
    from boxsdk.object.user import User
    from boxsdk.object.task_assignment import TaskAssignment
    from boxsdk.pagination.box_object_collection import BoxObjectCollection


class Task(BaseObject):
    """Represents a Box task."""
    _item_type = 'task'

    @api_call
    def assign(self, assignee: 'User') -> 'TaskAssignment':
        """
        Assign a task to a single user on a single file.

        :param assignee:
            The :class:`User` to assign the task to.
        :returns:
            A task assignment object.
        """
        url = self._session.get_url('task_assignments')
        body = {
            'task': {
                'type': 'task',
                'id': self.object_id,
            },
            'assign_to': {
                'id': assignee.object_id,
            },
        }
        response = self._session.post(url, data=json.dumps(body)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def assign_with_login(self, assignee_login: Optional[str]) -> 'TaskAssignment':
        """
        Used to assign a task to a single user with the login email address of the assignee.

        :param assignee_login:
            The login of the user to assign the task to.
        :returns:
            A task assignment object.
        """
        url = self._session.get_url('task_assignments')
        body = {
            'task': {
                'type': 'task',
                'id': self.object_id,
            },
            'assign_to': {
                'login': assignee_login,
            },
        }
        response = self._session.post(url, data=json.dumps(body)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_assignments(self, fields: Iterable[str] = None) -> 'BoxObjectCollection':
        """
        Get the entries in the file task assignment.

        :param fields:
            List of fields to request.
        :returns:
            An iterator of the entries in the file task assignment.
        """
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('assignments'),
            limit=None,
            marker=None,
            fields=fields,
            return_full_pages=False,
        )



from typing import Optional, Iterator, TYPE_CHECKING

from .box_object_collection import BoxObjectCollection

if TYPE_CHECKING:
    from boxsdk.session.session import Session


class LimitOffsetBasedObjectCollection(BoxObjectCollection):
    """
    An iterator of Box objects (BaseObjects) that were retrieved from a Box API endpoint that supports
    limit-offset type of pagination.

    See https://developer.box.com/en/guides/api-calls/pagination/ for more details.
    """

    def __init__(
            self,
            session: 'Session',
            url: str,
            limit: Optional[int] = None,
            fields: Iterator[str] = None,
            additional_params: Optional[dict] = None,
            return_full_pages: bool = False,
            offset: int = 0,
    ):
        """
        :param session:
            The Box session used to make requests.
        :param url:
            The endpoint url to hit.
        :param limit:
            The number of entries for each page to return. The default, as well as the upper limit of this value,
            differs by endpoint. See https://developer.box.com/en/reference. If limit is set to None, then the default
            limit (returned by Box in the response) is used.
        :param fields:
            List of fields to request. If None, will return the default fields for the object.
        :param additional_params:
            Additional HTTP params to send in the request.
        :param return_full_pages:
            If True, then the returned iterator for this collection will return full pages of Box objects on each
            call to next(). If False, the iterator will return a single Box object on each next() call.
        :param offset:
            The offset index to start paging from.
        """
        super().__init__(
            session,
            url,
            limit=limit,
            fields=fields,
            additional_params=additional_params,
            return_full_pages=return_full_pages,
        )
        self._offset = offset

    def _update_pointer_to_next_page(self, response_object: dict) -> None:
        """Baseclass override."""
        total_count = response_object['total_count']

        if 'limit' in response_object:
            self._limit, old_limit = int(response_object['limit']), self._limit

            # The API might use a lower limit than the client asked for, if the
            # client asked for a limit above the maximum limit for that endpoint.
            # The API is supposed to respond with the limit that it actually used.
            # If that is given, then use that limit for the offset calculation, and
            # also for the remainder of the paging.

            # Do not apply this same logic to "offset". Offset is not documented to be
            # changed in the response, so respecting that value can lead to undefined
            # behavior.

            # If the API erroneously sends a bad value for limit, we want to
            # avoid getting into an infinite chain of API calls. So abort with
            # a runtime error.
            if self._limit <= 0 < old_limit:
                self._offset = total_count  # Disable additional paging.
                raise RuntimeError(f'API returned limit={self._limit}, cannot continue paging')

        # de-none-ify the _offset value so that the arthimatic below works
        self._offset = self._offset or 0

        if total_count >= self._offset + self._limit:
            self._offset += self._limit
        else:
            self._offset = total_count

    def _has_more_pages(self, response_object: dict) -> bool:
        """Baseclass override."""
        return self._offset < response_object['total_count']

    def _next_page_pointer_params(self) -> dict:
        """Baseclass override."""
        return {'offset': self._offset}

    def next_pointer(self) -> int:
        """Baseclass override."""
        return self._offset



from unittest.mock import NonCallableMock
import pytest

from boxsdk.object.cloneable import Cloneable
from boxsdk.util.api_call_decorator import api_call


@pytest.fixture
def api_call_result():
    return {'bar': 'ƒøø'}


@pytest.fixture(name='api_call_method')
def api_call_method_fixture(api_call_result):

    @api_call
    def api_call_method(self, *args, **kwargs):
        return self, args, kwargs, api_call_result

    return api_call_method


@pytest.fixture
def cloneable_subclass_with_api_call_method(api_call_method):
    api_call_method_fixture = api_call_method

    # pylint:disable=abstract-method
    class CloneableSubclass(Cloneable):
        api_call_method = api_call_method_fixture

    return CloneableSubclass


@pytest.fixture
def mock_cloneable(cloneable_subclass_with_api_call_method):

    # pylint:disable=abstract-method
    class MockCloneable(cloneable_subclass_with_api_call_method, NonCallableMock):
        pass

    return MockCloneable(spec_set=cloneable_subclass_with_api_call_method, name='Cloneable')


def test_api_call_is_decorator():

    @api_call
    def func():
        pass

    assert callable(func)
    assert hasattr(func, '__get__')


def test_api_call_decorated_function_must_be_a_method():

    @api_call
    def func():
        pass

    with pytest.raises(TypeError):
        func()


def test_api_call_decorated_method_must_be_a_cloneable_method():

    class NonCloneable:
        @api_call
        def func(self):
            pass

    obj = NonCloneable()
    with pytest.raises(TypeError):
        obj.func()


def test_api_call_decorated_method_must_be_bound_to_an_instance_of_the_owner(mock_cloneable, api_call_method):
    # pylint:disable=abstract-method
    class CloneableSubclass2(Cloneable):
        pass

    with pytest.raises(TypeError):
        api_call_method.__get__(mock_cloneable, CloneableSubclass2)


def test_api_call_decorated_method_returns_itself_when_bound_to_none(api_call_method, cloneable_subclass_with_api_call_method):
    assert api_call_method.__get__(None, Cloneable) is api_call_method
    assert not hasattr(api_call_method.__get__(None, Cloneable), '__self__')
    assert cloneable_subclass_with_api_call_method.api_call_method is api_call_method
    assert not hasattr(cloneable_subclass_with_api_call_method.api_call_method, '__self__')


def test_api_call_decorated_method_binds_to_instance(mock_cloneable, api_call_method):
    assert api_call_method.__get__(mock_cloneable, Cloneable) is not api_call_method
    assert api_call_method.__get__(mock_cloneable, Cloneable).__self__ is mock_cloneable
    assert mock_cloneable.api_call_method is not api_call_method
    assert mock_cloneable.api_call_method.__self__ is mock_cloneable


def test_api_call_decorated_method_delegates_to_wrapped_method(mock_cloneable, api_call_result):
    args = (1, 2, 'ƒøø', 'bar')
    kwargs = {'bar': 'ƒøø'}
    assert mock_cloneable.api_call_method(*args, **kwargs) == (mock_cloneable, args, kwargs, api_call_result)


def test_api_call_decorated_method_can_be_called_as_an_unbound_method_with_an_instance_as_the_first_argument(
        mock_cloneable,
        api_call_result,
        cloneable_subclass_with_api_call_method,
):
    args = (1, 2, 'ƒøø', 'bar')
    kwargs = {'bar': 'ƒøø'}
    api_call_method = cloneable_subclass_with_api_call_method.api_call_method
    assert api_call_method(mock_cloneable, *args, **kwargs) == (mock_cloneable, args, kwargs, api_call_result)



from typing import Any

from .base_object import BaseObject


class SignTemplate(BaseObject):
    """
    Represents a Sign Template used by Box Sign
    """
    _item_type = 'sign-template'

    def get_url(self, *args: Any) -> str:
        """
        Returns the url for this sign template.
        """
        return self._session.get_url('sign_templates', self._object_id, *args)



import json
from typing import Any

from .base_object import BaseObject
from ..util.text_enum import TextEnum
from ..util.api_call_decorator import api_call


class CascadePolicyConflictResolution(TextEnum):
    PRESERVE_EXISTING = 'none'
    OVERWRITE = 'overwrite'


class MetadataCascadePolicy(BaseObject):
    """Represents a metadata cascade policy, which applies folder metadata to files in that folder."""

    _item_type = 'metadata_cascade_policy'

    def get_url(self, *args: Any) -> str:
        """
        Base class override to account for the correct pluralization.

        """
        return self._session.get_url('metadata_cascade_policies', self.object_id, *args)

    @api_call
    def force_apply(self, conflict_resolution: CascadePolicyConflictResolution) -> bool:
        """
        Applies the metadata values on the folder to all files within the folder.  The conflict resolution
        parameter determines how conflicts when the same metadata template is already applied to a file will be
        handled; either the file's existing values or the folder values can be given precendence.

        :param conflict_resolution:
            How conflicting metadata values should be reolved
        :returns:
            Whether the force application succeeded.
        """
        url = self.get_url('apply')
        body = {
            'conflict_resolution': conflict_resolution,
        }
        response = self._session.post(url, data=json.dumps(body), expect_json_response=False)
        return response.ok



import base64
import configparser
import os
import json
from pathlib import Path
from typing import Optional

from boxsdk.auth.jwt_auth import JWTAuth
from boxsdk.client import Client


JWT_CONFIG_ENV_VAR_NAME = 'JWT_CONFIG_BASE_64'
ADMIN_USER_ID_ENV_VAR_NAME = 'ADMIN_USER_ID'
CURRENT_DIR_PATH = str(Path(os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))))
CONFIG_PATH = f'{CURRENT_DIR_PATH}/integration_tests.cfg'
CONFIG_PARSER = configparser.ConfigParser()
CONFIG_PARSER.read(CONFIG_PATH)


def get_jwt_config() -> JWTAuth:
    jwt_config = read_jwt_config_from_env_var() or read_jwt_config_from_file()

    if not jwt_config:
        raise RuntimeError(
            f'JWT config cannot be loaded. Missing environment variable: {JWT_CONFIG_ENV_VAR_NAME} or JWT config path.'
        )
    return jwt_config


def get_admin_user_id() -> str:
    admin_user_id = os.getenv(ADMIN_USER_ID_ENV_VAR_NAME) or CONFIG_PARSER["JWT"].get('AdminUserID')

    if not admin_user_id:
        raise RuntimeError(
            f'Unknown admin user id. Missing environment variable: {ADMIN_USER_ID_ENV_VAR_NAME} or value in {CONFIG_PATH}.'
        )

    return admin_user_id


def read_jwt_config_from_env_var() -> Optional[JWTAuth]:
    jwt_config_base64 = os.getenv(JWT_CONFIG_ENV_VAR_NAME)
    if not jwt_config_base64:
        return None
    jwt_config_str = base64.b64decode(jwt_config_base64)
    jwt_config_json = json.loads(jwt_config_str)
    return JWTAuth.from_settings_dictionary(jwt_config_json)


def read_jwt_config_from_file() -> Optional[JWTAuth]:
    jwt_config_file_path = CONFIG_PARSER["JWT"].get('ConfigFilePath')
    if not jwt_config_file_path:
        return None
    return JWTAuth.from_settings_file(jwt_config_file_path)


def create_client(jwt_config: JWTAuth):
    return Client(jwt_config)


def create_user_client(jwt_config: JWTAuth):
    admin_user_id = get_admin_user_id()
    jwt_config.authenticate_user(admin_user_id)
    return Client(jwt_config)


config = get_jwt_config()
CLIENT = create_client(config)
USER_CLIENT = create_user_client(config)



import json
from datetime import datetime, timedelta
from unittest import mock
from unittest.mock import Mock, MagicMock, call
from test.unit.object.conftest import mock_user  # pylint:disable=unused-import

import pytz
import pytest
from pytest_lazyfixture import lazy_fixture

from boxsdk import BoxOAuthException
from boxsdk.auth.ccg_auth import CCGAuth
from boxsdk.config import API


USER_SUBJECT_TYPE = 'user'
ENTERPRISE_SUBJECT_TYPE = 'enterprise'


@pytest.fixture(scope='function')
def successful_token_response(successful_token_json_response):
    successful_token_response_mock = MagicMock()
    response = successful_token_json_response.copy()
    del response['refresh_token']
    successful_token_response_mock.items = response
    successful_token_response_mock.json = Mock(return_value=response)
    successful_token_response_mock.access_token = response['access_token']
    successful_token_response_mock.ok = True
    successful_token_response_mock.content = json.dumps(response)
    successful_token_response_mock.status_code = 200
    return successful_token_response_mock


@pytest.fixture
def unsuccessful_ccg_response(box_datetime, status_code, error_description, include_date_header, error_code):
    headers = {'Date': box_datetime.strftime('%a, %d %b %Y %H:%M:%S %Z')} if include_date_header else {}
    unsuccessful_response = Mock(headers=headers)
    unsuccessful_response.json.return_value = {'error_description': error_description, 'error': error_code}
    unsuccessful_response.status_code = status_code
    unsuccessful_response.ok = False
    return unsuccessful_response


@pytest.fixture()
def ccg_auth(client_id, client_secret, ccg_user_id, ccg_enterprise_id) -> CCGAuth:
    auth = CCGAuth(client_id=client_id, client_secret=client_secret, user=ccg_user_id, enterprise_id=ccg_enterprise_id)
    auth._session.get_retry_after_time = MagicMock(return_value=0)  # pylint:disable=protected-access
    return auth


@pytest.fixture
def expect_auth_retry(status_code, error_description, include_date_header, error_code):
    return status_code == 400 and 'exp' in error_description and include_date_header and error_code == 'invalid_grant'


@pytest.fixture
def box_datetime():
    return datetime.now(tz=pytz.utc) - timedelta(100)


@pytest.mark.parametrize('enterprise_id', (lazy_fixture('mock_enterprise_id'), None))
@pytest.mark.parametrize('user, expected_normalized_user_id',
                         [
                             (lazy_fixture('mock_user'), lazy_fixture('mock_user_id')),
                             (lazy_fixture('mock_user_id'), lazy_fixture('mock_user_id')),
                             (None, None)
                         ])
def test_successful_create_ccg_auth_object_and_normalize_user_id(client_id, client_secret, enterprise_id, user, expected_normalized_user_id):
    auth = CCGAuth(client_id=client_id, client_secret=client_secret, enterprise_id=enterprise_id, user=user)
    assert auth._user_id == expected_normalized_user_id  # pylint:disable=protected-access


def test_throw_type_error_when_invalid_user_object_provided(client_id, client_secret, mock_enterprise_id):
    with pytest.raises(TypeError):
        CCGAuth(
            client_id=client_id,
            client_secret=client_secret,
            enterprise_id=mock_enterprise_id,
            user=123123
        )


@pytest.mark.parametrize('ccg_user_id, ccg_enterprise_id, user',
                         [
                             (lazy_fixture('mock_user_id'), None, None),
                             (None, None, lazy_fixture('mock_user_id')),
                             (None, None, lazy_fixture('mock_user'))
                         ])
def test_authenticate_user_success(ccg_auth, user, mock_user_id):
    with mock.patch.object(CCGAuth, '_fetch_access_token', return_value='new_token') as fetch_access_token_mock:
        ccg_auth.authenticate_user(user)

        fetch_access_token_mock.assert_called_once_with(mock_user_id, USER_SUBJECT_TYPE, None)
        assert ccg_auth._user_id == mock_user_id  # pylint:disable=protected-access


@pytest.mark.parametrize('ccg_user_id', (None,))
@pytest.mark.parametrize('ccg_enterprise_id', (None,))
def test_authenticate_user_throws_error_when_user_not_provided(ccg_auth):
    with pytest.raises(ValueError):
        ccg_auth.authenticate_user(None)


@pytest.mark.parametrize('ccg_user_id, ccg_enterprise_id, enterprise_id',
                         [
                             (None, lazy_fixture('mock_enterprise_id'), None),
                             (None, None, lazy_fixture('mock_enterprise_id'))
                         ])
def test_authenticate_enterprise(ccg_auth, enterprise_id, mock_enterprise_id):
    with mock.patch.object(CCGAuth, '_fetch_access_token', return_value='new_token') as fetch_access_token_mock:
        ccg_auth.authenticate_instance(enterprise_id)

        fetch_access_token_mock.assert_called_once_with(mock_enterprise_id, ENTERPRISE_SUBJECT_TYPE, None)
        assert ccg_auth._enterprise_id == mock_enterprise_id  # pylint:disable=protected-access
        assert ccg_auth._user_id is None  # pylint:disable=protected-access


@pytest.mark.parametrize('ccg_user_id, ccg_enterprise_id, enterprise_id',
                         [
                             (None, None, None),
                             (None, lazy_fixture('mock_enterprise_id'), 'other_enterprise_300')
                         ])
def test_authenticate_user_throws_error_when_enetrprise_not_provided_or_conflicts(ccg_auth, enterprise_id):
    with pytest.raises(ValueError):
        ccg_auth.authenticate_instance(enterprise_id)


@pytest.mark.parametrize('ccg_user_id', (lazy_fixture('mock_user_id'),))
@pytest.mark.parametrize('ccg_enterprise_id', (lazy_fixture('mock_enterprise_id'),))
@pytest.mark.parametrize('status_code', (400, 401))
@pytest.mark.parametrize('error_description', ('invalid box_sub_type claim', 'invalid kid', "check the 'exp' claim"))
@pytest.mark.parametrize('error_code', ('invalid_grant', 'bad_request'))
@pytest.mark.parametrize('include_date_header', (True, False))
def test_auth_retry_for_invalid_exp_claim(
        ccg_auth,
        status_code,
        expect_auth_retry,
        unsuccessful_ccg_response,
        box_datetime,
        mock_enterprise_id
):
    side_effect = [BoxOAuthException(status_code, network_response=unsuccessful_ccg_response), 'jwt_token']
    with mock.patch.object(CCGAuth, '_fetch_access_token', return_value='new_token', side_effect=side_effect) as fetch_access_token_mock:

        if not expect_auth_retry:
            with pytest.raises(BoxOAuthException):
                ccg_auth.authenticate_instance()
        else:
            ccg_auth.authenticate_instance()

        expected_calls = [call(mock_enterprise_id, 'enterprise', None)]
        if expect_auth_retry:
            expected_calls.append(call(mock_enterprise_id, 'enterprise', box_datetime.replace(microsecond=0, tzinfo=None)))
        assert len(fetch_access_token_mock.mock_calls) == len(expected_calls)
        fetch_access_token_mock.assert_has_calls(expected_calls)


@pytest.mark.parametrize('ccg_user_id', (lazy_fixture('mock_user_id'),))
@pytest.mark.parametrize('ccg_enterprise_id', (lazy_fixture('mock_enterprise_id'),))
@pytest.mark.parametrize('status_code, error_description, error_code',
                         [(429, 'Request rate limit exceeded', 'rate_limit_exceeded'),
                          (500, 'Internal Server Error', 'internal_server_error')])
@pytest.mark.parametrize('include_date_header', (False,))
def test_auth_retry_for_rate_limit_and_server_errors(
        unsuccessful_ccg_response,
        ccg_auth,
        status_code,
        mock_enterprise_id
):

    side_effect = []
    expected_calls = []
    # Retries multiple times, but less than max retries. Then succeeds when it gets a token.
    for _ in range(API.MAX_RETRY_ATTEMPTS - 2):
        side_effect.append(BoxOAuthException(status_code, network_response=unsuccessful_ccg_response))
        expected_calls.append(call(mock_enterprise_id, ENTERPRISE_SUBJECT_TYPE, None))
    side_effect.append('jwt_token')
    expected_calls.append(call(mock_enterprise_id, 'enterprise', None))
    with mock.patch.object(CCGAuth, '_fetch_access_token', return_value='new_token', side_effect=side_effect) as fetch_access_token_mock:

        ccg_auth.authenticate_instance(mock_enterprise_id)

        assert len(fetch_access_token_mock.mock_calls) == len(expected_calls)
        fetch_access_token_mock.assert_has_calls(expected_calls)


@pytest.mark.parametrize('ccg_user_id', (lazy_fixture('mock_user_id'),))
@pytest.mark.parametrize('ccg_enterprise_id', (lazy_fixture('mock_enterprise_id'),))
@pytest.mark.parametrize('status_code, error_description, error_code',
                         [(429, 'Request rate limit exceeded', 'rate_limit_exceeded'),
                          (500, 'Internal Server Error', 'internal_server_error')])
@pytest.mark.parametrize('include_date_header', (False,))
def test_auth_max_retries_for_rate_limit_and_server_errors(
        unsuccessful_ccg_response,
        ccg_auth,
        status_code,
        mock_enterprise_id
):

    side_effect = []
    expected_calls = []
    # Retries multiple times, but less than max retries. Then succeeds when it gets a token.
    for _ in range(API.MAX_RETRY_ATTEMPTS + 1):
        side_effect.append(BoxOAuthException(status_code, network_response=unsuccessful_ccg_response))
        expected_calls.append(call(mock_enterprise_id, ENTERPRISE_SUBJECT_TYPE, None))
    with mock.patch.object(CCGAuth, '_fetch_access_token', return_value='new_token', side_effect=side_effect) as fetch_access_token_mock:

        with pytest.raises(BoxOAuthException) as error:
            ccg_auth.authenticate_instance(mock_enterprise_id)

        assert error.value.status == status_code
        assert len(fetch_access_token_mock.mock_calls) == len(expected_calls)
        fetch_access_token_mock.assert_has_calls(expected_calls)


@pytest.mark.parametrize('ccg_user_id', (lazy_fixture('mock_user_id'),))
@pytest.mark.parametrize('ccg_enterprise_id', (lazy_fixture('mock_enterprise_id'),))
def test_extract_token_from_success_response(
        successful_token_response,
        successful_token_json_response,
        ccg_auth,
        mock_enterprise_id
):
    with mock.patch.object(CCGAuth, '_execute_token_request', return_value=successful_token_response):

        ccg_auth.authenticate_instance(mock_enterprise_id)

        assert ccg_auth.access_token == successful_token_json_response['access_token']


@pytest.mark.parametrize('ccg_user_id', (lazy_fixture('mock_user_id'),))
@pytest.mark.parametrize('ccg_enterprise_id', (None,))
def test_refresh_client_authentication_when_client_id_is_provided(ccg_auth):
    ccg_auth.authenticate_instance = Mock()
    ccg_auth.authenticate_user = Mock()

    ccg_auth.refresh("expired_token")

    ccg_auth.authenticate_user.assert_called_once()
    ccg_auth.authenticate_instance.assert_not_called()


@pytest.mark.parametrize('ccg_user_id', (None,))
@pytest.mark.parametrize('ccg_enterprise_id', (lazy_fixture('mock_enterprise_id'),))
def test_refresh_enterprise_authentication_when_client_id_is_not_provided(ccg_auth):
    ccg_auth.authenticate_instance = Mock()
    ccg_auth.authenticate_user = Mock()

    ccg_auth.refresh("expired_token")

    ccg_auth.authenticate_user.assert_not_called()
    ccg_auth.authenticate_instance.assert_called_once()


@pytest.mark.parametrize('ccg_user_id', (lazy_fixture('mock_user_id'),))
@pytest.mark.parametrize('ccg_enterprise_id', (lazy_fixture('mock_enterprise_id'),))
@pytest.mark.parametrize('subject_id, subject_type',
                         [
                             (lazy_fixture('mock_user_id'), USER_SUBJECT_TYPE),
                             (lazy_fixture('mock_enterprise_id'), ENTERPRISE_SUBJECT_TYPE)
                         ])
def test_send_authentication_request_with_correct_params(
        ccg_auth,
        client_id,
        client_secret,
        subject_id,
        subject_type
):
    expected_data = {
        'grant_type': 'client_credentials',
        'client_id': client_id,
        'client_secret': client_secret,
        'box_subject_id': subject_id,
        'box_subject_type': subject_type
    }
    ccg_auth.send_token_request = Mock(return_value=('new_token', None))

    # pylint:disable=protected-access
    ccg_auth._fetch_access_token(subject_id, subject_type, None)

    ccg_auth.send_token_request.assert_called_once_with(expected_data, access_token=None, expect_refresh_token=False)



from boxsdk.config import API
from boxsdk.object.file_version_retention import FileVersionRetention


def test_get(test_file_version_retention, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/file_version_retentions/{test_file_version_retention.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'file_version_retention',
        'id': test_file_version_retention.object_id,
    }
    file_version_retention = test_file_version_retention.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(file_version_retention, FileVersionRetention)



import json

from boxsdk.object.terms_of_service_user_status import TermsOfServiceUserStatus
from boxsdk.config import API


def test_get(test_terms_of_service_user_status, mock_box_session):
    created_at = '2016-05-18T17:38:03-07:00'
    expected_url = f'{API.BASE_API_URL}/terms_of_service_user_statuses/{test_terms_of_service_user_status.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'terms_of_service_user_status',
        'id': test_terms_of_service_user_status.object_id,
        'created_at': created_at
    }
    terms_of_service = test_terms_of_service_user_status.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(terms_of_service, TermsOfServiceUserStatus)
    assert terms_of_service.type == test_terms_of_service_user_status.object_type
    assert terms_of_service.id == test_terms_of_service_user_status.object_id
    assert terms_of_service.created_at == created_at


def test_accept(test_terms_of_service_user_status, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/terms_of_service_user_statuses/{test_terms_of_service_user_status.object_id}'
    mock_box_session.put.return_value.json.return_value = {
        'type': test_terms_of_service_user_status.object_type,
        'id': test_terms_of_service_user_status.object_id,
        'is_accepted': True
    }
    data = {'is_accepted': True}
    terms_of_service_user_status = test_terms_of_service_user_status.accept()
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=None, params=None)
    assert isinstance(terms_of_service_user_status, TermsOfServiceUserStatus)
    assert terms_of_service_user_status.type == test_terms_of_service_user_status.object_type
    assert terms_of_service_user_status.id == test_terms_of_service_user_status.object_id
    assert terms_of_service_user_status.is_accepted is True


def test_reject(test_terms_of_service_user_status, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/terms_of_service_user_statuses/{test_terms_of_service_user_status.object_id}'
    mock_box_session.put.return_value.json.return_value = {
        'type': 'terms_of_service_user_status',
        'id': test_terms_of_service_user_status.object_id,
        'is_accepted': False
    }
    data = {'is_accepted': False}
    terms_of_service_user_status = test_terms_of_service_user_status.reject()
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=None, params=None)
    assert isinstance(terms_of_service_user_status, TermsOfServiceUserStatus)
    assert terms_of_service_user_status.type == test_terms_of_service_user_status.object_type
    assert terms_of_service_user_status.id == test_terms_of_service_user_status.object_id
    assert terms_of_service_user_status.is_accepted is False



from datetime import datetime

import pytest

from boxsdk import BoxAPIException
from test.integration_new import CLIENT
from test.integration_new.context_managers.box_sign_request import BoxTestSignRequest
from test.integration_new.context_managers.box_test_file import BoxTestFile
from test.integration_new.context_managers.box_test_folder import BoxTestFolder

SIGN_REQUEST_TESTS_DIRECTORY_NAME = 'sign-request-integration-tests'


@pytest.fixture(scope="module", autouse=True)
def parent_folder():
    with BoxTestFolder(name=f'{SIGN_REQUEST_TESTS_DIRECTORY_NAME} {datetime.now()}') as folder:
        yield folder


def test_webhook_sign_request(parent_folder, small_file_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as test_file:
        file = {
            'id': test_file.object_id,
            'type': test_file.object_type,
        }
        files = [file]
        signer = {
            'email': 'signer@mail.com',
        }
        signers = [signer]

        with BoxTestSignRequest(files=files, signers=signers, parent_folder_id=parent_folder.id) as sign_request:
            webhook_url = 'https://example.com/webhook'
            new_webhook_url = 'https://example.com/new-webhook'

            assert len(sign_request.sign_files['files']) == 1

            sign_file = sign_request.sign_files['files'][0]
            webhook_file = CLIENT.file(file_id=sign_file['id'])
            webhook = CLIENT.create_webhook(
                webhook_file,
                ['SIGN_REQUEST.COMPLETED', 'SIGN_REQUEST.DECLINED', 'SIGN_REQUEST.EXPIRED'],
                webhook_url
            )

            try:
                assert webhook.id
                assert webhook.triggers == ['SIGN_REQUEST.COMPLETED', 'SIGN_REQUEST.DECLINED', 'SIGN_REQUEST.EXPIRED']
                assert webhook.address == webhook_url

                update_data = {
                    'address': new_webhook_url,
                    'triggers': ['SIGN_REQUEST.COMPLETED']
                }
                webhook = webhook.update_info(data=update_data)

                assert webhook.address == new_webhook_url
                assert webhook.triggers == ['SIGN_REQUEST.COMPLETED']
            finally:
                webhook.delete()

            with pytest.raises(BoxAPIException):
                webhook.get()



from functools import partial
from io import IOBase
from numbers import Number
from unittest.mock import MagicMock, Mock, PropertyMock, call, patch, ANY
from requests.exceptions import RequestException, SSLError, ConnectionError as RequestsConnectionError

import pytest

from boxsdk import CCGAuth
from boxsdk.auth.oauth2 import OAuth2
from boxsdk.config import API, Proxy
from boxsdk.exception import BoxAPIException, BoxException
from boxsdk.network.default_network import DefaultNetwork, DefaultNetworkResponse
from boxsdk.session.box_response import BoxResponse
from boxsdk.session.session import Session, Translator, AuthorizedSession


@pytest.fixture(scope='function', params=[False, True])
def translator(default_translator, request):  # pylint:disable=unused-argument
    if request.param:
        return Translator(extend_default_translator=True, new_child=True)
    return None


@pytest.fixture
def initial_access_token():
    return 'fake_access_token'


@pytest.fixture
def mock_oauth(initial_access_token):
    mock_oauth = MagicMock(OAuth2)
    mock_oauth.access_token = initial_access_token
    return mock_oauth


@pytest.fixture
def mock_network_layer():
    return Mock(DefaultNetwork)


@pytest.fixture
def unauthorized_session(mock_network_layer, translator):
    # pylint:disable=redefined-outer-name
    return Session(network_layer=mock_network_layer, translator=translator)


@pytest.fixture
def box_session(mock_oauth, mock_network_layer, translator):
    # pylint:disable=redefined-outer-name
    return AuthorizedSession(oauth=mock_oauth, network_layer=mock_network_layer, translator=translator)


@pytest.fixture
def ccg_auth(client_id, client_secret, mock_user_id, mock_enterprise_id, box_session) -> CCGAuth:
    # pylint:disable=protected-access
    auth = CCGAuth(client_id=client_id, client_secret=client_secret, user=mock_user_id, enterprise_id=mock_enterprise_id)
    auth._session = box_session
    return auth


@pytest.mark.parametrize('test_method', [
    Session.get,
    Session.post,
    Session.put,
    Session.delete,
    Session.options,
])
def test_box_session_handles_unauthorized_response(
        test_method,
        box_session,
        mock_oauth,
        mock_network_layer,
        unauthorized_response,
        generic_successful_response,
        test_url,
):
    # pylint:disable=redefined-outer-name

    def get_access_token_from_auth_object():
        return mock_oauth.access_token

    mock_network_layer.request.side_effect = mock_responses = [unauthorized_response, generic_successful_response]
    for mock_response in mock_responses:
        type(mock_response).access_token_used = PropertyMock(side_effect=get_access_token_from_auth_object)

    def refresh(access_token_used):
        assert access_token_used == mock_oauth.access_token
        mock_oauth.access_token = 'fake_new_access_token'
        return (mock_oauth.access_token, None)

    mock_oauth.refresh.side_effect = refresh

    box_response = test_method(box_session, url=test_url)
    assert box_response.status_code == 200


@pytest.mark.parametrize('test_method', [
    Session.get,
    Session.post,
    Session.put,
    Session.delete,
    Session.options,
])
@pytest.mark.parametrize('initial_access_token', [None])
def test_box_session_gets_access_token_before_request(
        test_method,
        box_session,
        mock_oauth,
        mock_network_layer,
        generic_successful_response,
        test_url,
):
    # pylint:disable=redefined-outer-name

    def get_access_token_from_auth_object():
        return mock_oauth.access_token

    mock_network_layer.request.side_effect = mock_responses = [generic_successful_response]
    for mock_response in mock_responses:
        type(mock_response).access_token_used = PropertyMock(side_effect=get_access_token_from_auth_object)

    def refresh(access_token_used):
        assert access_token_used == mock_oauth.access_token
        mock_oauth.access_token = 'fake_new_access_token'
        return (mock_oauth.access_token, None)

    mock_oauth.refresh.side_effect = refresh

    box_response = test_method(box_session, url=test_url, auto_session_renewal=True)
    assert box_response.status_code == 200


@pytest.mark.parametrize('test_method', [
    Session.get,
    Session.post,
    Session.put,
    Session.delete,
    Session.options,
    partial(Session.request, method='head'),
])
def test_box_session_retries_response_after_retry_after(
        test_method,
        box_session,
        mock_network_layer,
        retry_after_response,
        generic_successful_response,
        test_url,
):
    # pylint:disable=redefined-outer-name
    try:
        if retry_after_response.status_code == 202 and test_method.__name__ == 'delete':
            pytest.xfail("Delete operation should not be retried on 202 status code")
    except AttributeError:
        pass

    mock_network_layer.request.side_effect = [retry_after_response, generic_successful_response]
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)

    with patch('random.uniform', return_value=0.68):
        box_response = test_method(box_session, url=test_url)
    assert box_response.status_code == 200
    assert len(mock_network_layer.retry_after.call_args_list) == 1
    assert isinstance(mock_network_layer.retry_after.call_args[0][0], Number)
    assert round(mock_network_layer.retry_after.call_args[0][0], 4) == 1


@pytest.mark.parametrize('test_method', [
    Session.get,
    Session.post,
    Session.put,
    Session.delete,
    Session.options,
    partial(Session.request, method='head'),
])
def test_box_session_retries_request_after_server_error(
        test_method,
        box_session,
        mock_network_layer,
        server_error_response,
        generic_successful_response,
        test_url,
):
    # pylint:disable=redefined-outer-name
    mock_network_layer.request.side_effect = [server_error_response, server_error_response, generic_successful_response]
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)

    with patch('random.uniform', return_value=0.68):
        box_response = test_method(box_session, url=test_url)
    assert box_response.status_code == 200
    assert box_response.json() == generic_successful_response.json()
    assert box_response.ok == generic_successful_response.ok
    assert box_response.content == generic_successful_response.content
    assert len(mock_network_layer.retry_after.call_args_list) == 2
    assert isinstance(mock_network_layer.retry_after.call_args_list[0][0][0], Number)
    assert isinstance(mock_network_layer.retry_after.call_args_list[1][0][0], Number)
    assert round(mock_network_layer.retry_after.call_args_list[0][0][0], 4) == 1.18
    assert round(mock_network_layer.retry_after.call_args_list[1][0][0], 4) == 2.36


def test_box_session_seeks_file_after_retry(box_session, mock_network_layer, server_error_response, generic_successful_response, test_url):
    # pylint:disable=redefined-outer-name
    mock_network_layer.request.side_effect = [server_error_response, generic_successful_response]
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)
    mock_file_1, mock_file_2 = MagicMock(IOBase), MagicMock(IOBase)
    mock_file_1.tell.return_value = 0
    mock_file_2.tell.return_value = 3
    files = {'file': ('unused', mock_file_1), 'f2': ('unused', mock_file_2)}

    box_response = box_session.post(url=test_url, files=files)
    assert box_response.status_code == 200
    assert box_response.json() == generic_successful_response.json()
    assert box_response.ok == generic_successful_response.ok
    mock_file_1.tell.assert_called_with()
    mock_file_2.tell.assert_called_with()
    mock_file_1.seek.assert_called_with(0)
    assert mock_file_1.seek.call_count == 2
    assert mock_file_1.seek.has_calls(call(0) * 2)
    mock_file_2.seek.assert_called_with(3)
    assert mock_file_2.seek.call_count == 2
    assert mock_file_2.seek.has_calls(call(3) * 2)


def test_box_session_raises_for_non_json_response(box_session, mock_network_layer, non_json_response, generic_successful_response, test_url):
    # pylint:disable=redefined-outer-name
    mock_network_layer.request.side_effect = [non_json_response, generic_successful_response]
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)

    box_session.get(url=test_url)


def test_box_session_raises_for_non_json_response_after_retry(box_session, mock_network_layer, non_json_response, test_url):
    # pylint:disable=redefined-outer-name
    mock_network_layer.request.side_effect = [non_json_response] * (API.MAX_RETRY_ATTEMPTS + 1)
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)

    with pytest.raises(BoxAPIException):
        box_session.get(url=test_url)


def test_box_session_raises_for_failed_response(box_session, mock_network_layer, bad_network_response, test_url):
    # pylint:disable=redefined-outer-name
    mock_network_layer.request.side_effect = [bad_network_response]

    with pytest.raises(BoxAPIException):
        box_session.get(url=test_url)


@pytest.mark.parametrize('exc_message', [
    'Connection aborted',
    "Connection broken: ConnectionResetError(54, 'Connection reset by peer')"
])
def test_box_session_retries_connection_aborted_exception(
        box_session, mock_network_layer, generic_successful_request_response, test_url, exc_message
):
    mock_network_layer.request.side_effect = [RequestsConnectionError(exc_message), generic_successful_request_response]
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)

    box_response = box_session.get(url=test_url)
    assert box_response.status_code == 200


def test_box_session_retries_connection_aborted_exception_on_ccg_auth_call(successful_token_response, ccg_auth, mock_user_id, access_token):
    # pylint:disable=protected-access
    ccg_auth._session._network_layer.request.side_effect = [RequestsConnectionError('Connection aborted'), successful_token_response]
    ccg_auth._session._network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)

    new_access_token = ccg_auth.authenticate_user(mock_user_id)
    assert new_access_token == access_token


def test_box_session_retries_requests_library_exceptions_only_once(box_session, mock_network_layer, test_url, generic_successful_request_response):
    mock_network_layer.request.side_effect = [
        RequestException('Connection aborted'),
        RequestException('Connection aborted'),
        generic_successful_request_response
    ]
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)

    with pytest.raises(RequestException):
        box_session.get(url=test_url)


def test_box_session_raises_requests_library_exception_when_set_no_retries(
        box_session, mock_network_layer, test_url, generic_successful_request_response
):
    API.MAX_RETRY_ATTEMPTS = 0
    try:
        mock_network_layer.request.side_effect = [RequestException('Connection aborted'), generic_successful_request_response]

        with pytest.raises(RequestException):
            box_session.get(url=test_url)
    finally:
        API.MAX_RETRY_ATTEMPTS = 5


def test_box_session_retries_and_reauthenticates_on_violation_of_protocol_exception(
        box_session, mock_oauth, mock_network_layer, generic_successful_request_response, test_url
):
    def refresh(access_token_used):
        assert access_token_used == mock_oauth.access_token
        mock_oauth.access_token = 'fake_new_access_token'
        return (mock_oauth.access_token, None)

    mock_network_layer.request.side_effect = [
        SSLError("SSLError(SSLEOFError(8, 'EOF occurred in violation of protocol (_ssl.c:2396)')))"),
        generic_successful_request_response
    ]
    mock_network_layer.retry_after.side_effect = lambda delay, request, *args, **kwargs: request(*args, **kwargs)
    mock_oauth.refresh.side_effect = refresh

    box_session.get(url=test_url)

    assert mock_network_layer.request.call_count == 2
    assert mock_network_layer.request.mock_calls[0][2]['access_token'] == 'fake_access_token'
    assert mock_network_layer.request.mock_calls[1][2]['access_token'] == 'fake_new_access_token'


def test_box_session_does_not_retry_other_requests_library_exceptions_than_specified(box_session, mock_network_layer, test_url):
    mock_network_layer.request.side_effect = [RequestException('Unknown error')]

    with pytest.raises(RequestException):
        box_session.get(url=test_url)


def test_box_session_raises_for_failed_response_with_error_and_error_description(box_session, mock_network_layer, bad_network_response_400, test_url):
    mock_network_layer.request.side_effect = [bad_network_response_400]
    try:
        box_session.get(url=test_url)
        pytest.fail('Should throw exception because of bad network response')
    except BoxAPIException as exception:
        assert exception.code == 'Example Error'
        assert exception.message == 'Example Error Description'


def test_box_session_raises_for_failed_non_json_response(box_session, mock_network_layer, failed_non_json_response, test_url):
    # pylint:disable=redefined-outer-name
    mock_network_layer.request.side_effect = [failed_non_json_response]

    with pytest.raises(BoxAPIException):
        box_session.get(url=test_url, expect_json_response=False)


def test_box_response_properties_pass_through_to_network_response_properties():
    mock_network_response = Mock(DefaultNetworkResponse)
    box_result = BoxResponse(mock_network_response)
    assert box_result.json() == mock_network_response.json()
    assert box_result.content == mock_network_response.content
    assert box_result.ok == mock_network_response.ok
    assert box_result.status_code == mock_network_response.status_code
    assert box_result.network_response == mock_network_response


def test_translator(box_session, translator, default_translator, original_default_translator):
    assert isinstance(box_session.translator, Translator)
    assert box_session.translator == default_translator
    if translator:
        assert box_session.translator is translator

    # Test that adding new registrations works.

    class Foo:
        pass

    item_type = 'ƒøø'
    box_session.translator.register(item_type, Foo)
    assert box_session.translator.get(item_type) is Foo

    # Test that adding new registrations does not affect global state.
    assert default_translator == original_default_translator
    assert (set(box_session.translator) - set(default_translator)) == {item_type}


def test_session_uses_global_config(box_session, mock_network_layer, generic_successful_response, monkeypatch):
    mock_network_layer.request.side_effect = generic_successful_response
    example_dot_com = 'https://example.com/'
    monkeypatch.setattr(API, 'BASE_API_URL', example_dot_com)
    assert example_dot_com in box_session.get_url('foo', 'bar')


def test_session_uses_local_config(box_session, mock_network_layer, generic_successful_response, monkeypatch):
    mock_network_layer.request.side_effect = generic_successful_response
    example_dot_com = 'https://example.com/'
    box_session.api_config.BASE_API_URL = example_dot_com
    monkeypatch.setattr(API, 'BASE_API_URL', 'https://api.box.com')
    assert example_dot_com in box_session.get_url('foo', 'bar')


@pytest.mark.parametrize(
    'attempt_number,retry_after_header,expected_result',
    [
        (0, '', 1.18),
        (1, '', 2.36),
        (2, '', 4.72),
        (3, '', 9.44),
        (4, '', 18.88),
    ]
)
def test_get_retry_after_time(box_session, attempt_number, retry_after_header, expected_result):
    with patch('random.uniform', return_value=0.68):
        retry_time = box_session.get_retry_after_time(attempt_number, retry_after_header)  # pylint: disable=protected-access
    retry_time = round(retry_time, 4)
    assert retry_time == expected_result


@pytest.mark.parametrize(
    'test_proxy_url,test_proxy_auth,expected_proxy_dict',
    [
        ('http://example-proxy.com', {'user': 'test_user', 'password': 'test_password', },
         {'http': 'http://test_user:test_password@example-proxy.com', 'https': 'http://test_user:test_password@example-proxy.com'}),
        ('http://example-proxy.com', None, {'http': 'http://example-proxy.com', 'https': 'http://example-proxy.com'}),
    ]
)
def test_proxy_attaches_to_request_correctly(
        box_session,
        monkeypatch,
        mock_network_layer,
        generic_successful_response,
        test_proxy_url, test_proxy_auth,
        expected_proxy_dict):
    monkeypatch.setattr(Proxy, 'URL', test_proxy_url)
    monkeypatch.setattr(Proxy, 'AUTH', test_proxy_auth)
    mock_network_layer.request.side_effect = [generic_successful_response]
    box_session.request('GET', test_proxy_url)
    mock_network_layer.request.assert_called_once_with(
        'GET',
        test_proxy_url,
        access_token='fake_access_token',
        headers=ANY,
        proxies=expected_proxy_dict,
        log_response_content=True,
    )


def test_proxy_malformed_dict_does_not_attach(box_session, monkeypatch, mock_network_layer, generic_successful_response):
    test_proxy_url = 'http://example.com'
    test_proxy_auth = {
        'foo': 'bar',
    }
    monkeypatch.setattr(Proxy, 'URL', test_proxy_url)
    monkeypatch.setattr(Proxy, 'AUTH', test_proxy_auth)
    mock_network_layer.request.side_effect = [generic_successful_response]
    with pytest.raises(BoxException) as exc_info:
        box_session.request('GET', test_proxy_url)
    assert isinstance(exc_info.value, BoxException)
    assert exc_info.value.args[0] == "The proxy auth dict you provided does not match pattern " \
                                     "{'user': 'example_user', 'password': 'example_password'}"


def test_proxy_network_config_property(box_session):
    assert isinstance(box_session.proxy_config, Proxy)



from boxsdk.util.text_enum import TextEnum
from .base_object import BaseObject


class ResolutionState(TextEnum):
    """An enum of possible resolution states"""
    COMPLETED = 'completed'
    INCOMPLETE = 'incomplete'
    APPROVED = 'approved'
    REJECTED = 'rejected'


class TaskAssignment(BaseObject):
    """Represents a Box task."""

    _item_type = 'task_assignment'



from datetime import datetime

import pytest

from boxsdk.object.collaboration import CollaborationRole

from test.integration_new.context_managers.box_test_file import BoxTestFile
from test.integration_new.context_managers.box_test_folder import BoxTestFolder

FILE_TESTS_DIRECTORY_NAME = 'collaboration-integration-tests'


@pytest.fixture(scope="module", autouse=True)
def parent_folder():
    with BoxTestFolder(name=f'{FILE_TESTS_DIRECTORY_NAME} {datetime.now()}') as folder:
        yield folder


def test_collaboration(parent_folder, small_file_path, other_user,):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as test_file:
        collaboration = test_file.collaborate(other_user, CollaborationRole.VIEWER)
        try:
            assert collaboration.item.id == test_file.id
            assert collaboration.accessible_by.id == other_user.id
            assert collaboration.role == CollaborationRole.VIEWER

            updated_expiration_date = '2088-01-01T00:00:00-08:00'
            collaboration_update = {'role': CollaborationRole.EDITOR, 'expires_at': updated_expiration_date}
            updated_collaboration = collaboration.update_info(data=collaboration_update)

            assert updated_collaboration.role == CollaborationRole.EDITOR
            assert updated_collaboration.expires_at != collaboration.expires_at
            assert updated_collaboration.expires_at == updated_expiration_date
        finally:
            collaboration.delete()



import json

from typing import TYPE_CHECKING, Any, Union

from .base_object import BaseObject
from ..exception import BoxValueError
from ..util.api_call_decorator import api_call
from ..util.datetime_formatter import normalize_date_to_rfc3339_format
from ..util.default_arg_value import SDK_VALUE_NOT_SET

if TYPE_CHECKING:
    from boxsdk.object.folder import Folder
    from boxsdk.object.collection import Collection


class BaseItem(BaseObject):

    @api_call
    def copy(self, *, parent_folder: 'Folder', name: str = None, **_kwargs) -> 'BaseItem':
        """Copy the item to the given folder.

        :param parent_folder:
            The folder to which the item should be copied.
        :param name:
            A new name for the item, in case there is already another item in the new parent folder with the same name.
        """
        self.validate_item_id(self._object_id)
        url = self.get_url('copy')
        data = {
            'parent': {'id': parent_folder.object_id}
        }
        if name is not None:
            data['name'] = name
        box_response = self._session.post(url, data=json.dumps(data))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def move(self, parent_folder: 'Folder', name: str = None) -> 'BaseItem':
        """
        Move the item to the given folder.

        :param parent_folder:
            The parent `Folder` object, where the item will be moved to.
        :param name:
            A new name for the item, in case there is already another item in the new parent folder with the same name.
        """
        data = {
            'parent': {'id': parent_folder.object_id}
        }
        if name is not None:
            data['name'] = name
        return self.update_info(data=data)

    @api_call
    def rename(self, name: str) -> 'BaseItem':
        """
        Rename the item to a new name.

        :param name:
            The new name, you want the item to be renamed to.
        """
        data = {
            'name': name,
        }
        return self.update_info(data=data)

    @api_call
    def create_shared_link(self, **kwargs: Any) -> Any:
        """
        Create a shared link for the item with the given access permissions.

        :param kwargs:
            Keyword arguments passed from overriding method used to build request properties.
        :return:
            The updated object with shared link.
            Returns a new object of the same type, without modifying the original object passed as self.
        """
        shared_link = {}

        if kwargs.get('access') is not None:
            shared_link['access'] = kwargs.get('access')

        if kwargs.get('unshared_at') is not SDK_VALUE_NOT_SET:
            shared_link['unshared_at'] = normalize_date_to_rfc3339_format(kwargs.get('unshared_at'))

        permissions = {}
        if kwargs.get('allow_download') is not None:
            permissions['can_download'] = kwargs.get('allow_download')
        if kwargs.get('allow_preview') is not None:
            permissions['can_preview'] = kwargs.get('allow_preview')
        if kwargs.get('allow_edit') is not None:
            permissions['can_edit'] = kwargs.get('allow_edit')
        if permissions:
            shared_link['permissions'] = permissions

        if kwargs.get('password') is not None:
            shared_link['password'] = kwargs.get('password')

        if kwargs.get('vanity_name') is not None:
            shared_link['vanity_name'] = kwargs.get('vanity_name')

        data = {'shared_link': shared_link}
        update_info_kwargs = {'etag': kwargs.get('etag')} if kwargs.get('etag') is not None else {}

        return self.update_info(data=data, **update_info_kwargs)

    @api_call
    def get_shared_link(self, **kwargs: Any) -> str:
        """
        Get a shared link for the item with the given access permissions.
        This url leads to a Box.com shared link page, where the item can be previewed, downloaded, etc.

        :param kwargs:
            Keyword arguments passed from overriding method used to create a new shared link.
        :returns:
            The URL of the shared link.
        """
        item = self.create_shared_link(**kwargs)
        return item.shared_link['url']  # pylint:disable=no-member

    @api_call
    def remove_shared_link(self, **kwargs: Any) -> bool:
        """
        Delete the shared link for the item.

        :param kwargs:
            Keyword arguments passed from overriding method used to build request properties.
        :returns:
            Whether or not the update was successful.
        """
        data = {'shared_link': None}
        update_info_kwargs = {'etag': kwargs.get('etag')} if kwargs.get('etag') is not None else {}

        item = self.update_info(data=data, **update_info_kwargs)
        return item.shared_link is None  # pylint:disable=no-member

    @api_call
    def add_to_collection(self, collection: 'Collection') -> 'BaseItem':
        """
        Add the item to a collection. This method is not currently safe from race conditions.

        :param collection:
            The collection to add the item to.
        :return:
            This item.
        """
        collections = self.get(fields=['collections']).collections  # pylint:disable=no-member
        collections.append({'id': collection.object_id})
        data = {
            'collections': collections
        }
        return self.update_info(data=data)

    @api_call
    def remove_from_collection(self, collection: 'Collection') -> 'BaseItem':
        """
        Remove the item from a collection.  This method is not currently safe from race conditions.

        :param collection:
            The collection to remove the item from.
        :return:
            This item.
        """
        collections = self.get(fields=['collections']).collections  # pylint:disable=no-member
        updated_collections = [c for c in collections if c['id'] != collection.object_id]
        data = {
            'collections': updated_collections
        }
        return self.update_info(data=data)

    @staticmethod
    def validate_item_id(item_id: Union[str, int]) -> None:
        """
        Validates an item ID is numeric

        :param item_id
        :raises:
            BoxException: if item_id is not numeric
        :returns:
        """
        if not isinstance(item_id, int) and not item_id.isdigit():
            raise BoxValueError("Invalid item ID")



import json
from typing import Any, Union, TYPE_CHECKING, Optional, Iterable

from .base_object import BaseObject
from ..pagination.marker_based_object_collection import MarkerBasedObjectCollection
from ..util.api_call_decorator import api_call

if TYPE_CHECKING:
    from boxsdk.object.user import User
    from boxsdk.object.folder import Folder
    from boxsdk.object.file import File
    from boxsdk.object.file_version import FileVersion
    from boxsdk.object.legal_hold_policy_assignment import LegalHoldPolicyAssignment
    from boxsdk.pagination.box_object_collection import BoxObjectCollection


class LegalHoldPolicy(BaseObject):
    """Represents a Box legal_hold_policy"""

    _item_type = 'legal_hold_policy'

    def get_url(self, *args: Any) -> str:
        return self._session.get_url('legal_hold_policies', self._object_id, *args)

    @api_call
    def assign(self, assignee: Union['FileVersion', 'File', 'Folder', 'User']) -> 'LegalHoldPolicyAssignment':
        """Assign legal hold policy

        :param assignee:
            The `file_version`, `file`, `folder`, or `user` to assign the legal hold policy to.
        :returns:
            A legal hold policy assignment object
        """
        url = self._session.get_url('legal_hold_policy_assignments')
        body = {
            'policy_id': self.object_id,
            'assign_to': {
                'type': assignee.object_type,
                'id': assignee.object_id
            }
        }
        response = self._session.post(url, data=json.dumps(body)).json()
        return self.translator.translate(
            self._session,
            response,
        )

    @api_call
    def get_assignments(
            self,
            assign_to_type: Optional[str] = None,
            assign_to_id: Optional[str] = None,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the legal hold policy assignment using limit-offset paging.

        :param assign_to_type:
            Filter assignments of this type only. Can be `file_version`, `file`, `folder`, or `user`
        :param assign_to_id:
            Filter assignments to this ID only
        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from
        :param fields:
            List of fields to request
        :returns:
            An iterator of the entries in the legal hold policy assignment
        """
        additional_params = {
            'policy_id': self.object_id,
        }
        if assign_to_type is not None:
            additional_params['assign_to_type'] = assign_to_type
        if assign_to_id is not None:
            additional_params['assign_to_id'] = assign_to_id
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self._session.get_url('legal_hold_policy_assignments'),
            additional_params=additional_params,
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    @api_call
    def get_file_version_legal_holds(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get legal holds for a file version.

        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from
        :param fields:
            List of fields to request
        :returns:
            An iterator of the entries in the file version legal holds
        """
        additional_params = {
            'policy_id': self.object_id,
        }
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self._session.get_url('file_version_legal_holds'),
            additional_params=additional_params,
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )



import json
import pytest

from boxsdk.config import API
from boxsdk.object.collaboration import CollaborationRole, CollaborationStatus


@pytest.mark.parametrize('data', [
    {},
    {'role': CollaborationRole.EDITOR},
    {'role': CollaborationRole.VIEWER},
    {'status': CollaborationStatus.ACCEPTED},
    {'status': CollaborationStatus.REJECTED},
    {'role': CollaborationRole.EDITOR, 'status': CollaborationStatus.ACCEPTED},
])
def test_update_info_returns_the_correct_response(
        test_collaboration,
        mock_box_session,
        mock_collab_response,
        data):
    # pylint:disable=protected-access
    expected_url = test_collaboration.get_url()
    mock_box_session.put.return_value = mock_collab_response
    update_response = test_collaboration.update_info(**data)
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(data),
        headers=None,
        params=None,
    )
    assert isinstance(update_response, test_collaboration.__class__)
    assert update_response.object_id == test_collaboration.object_id


@pytest.mark.parametrize('data', [
    {},
    {'role': CollaborationRole.EDITOR, 'status': CollaborationStatus.REJECTED},
    {'role': CollaborationRole.EDITOR, 'status': CollaborationStatus.ACCEPTED},
    {'role': CollaborationRole.EDITOR, 'expires_at': '2025-08-29T23:59:00-07:00'},
    {'role': CollaborationRole.EDITOR, 'can_view_path': True},
])
def test_update_info_returns_the_correct_response_with_data_param(
        test_collaboration,
        mock_box_session,
        mock_collab_response,
        data):
    # pylint:disable=protected-access
    expected_url = test_collaboration.get_url()
    mock_box_session.put.return_value = mock_collab_response
    update_response = test_collaboration.update_info(data=data)
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(data),
        headers=None,
        params=None,
    )
    assert isinstance(update_response, test_collaboration.__class__)
    assert update_response.object_id == test_collaboration.object_id


def test_update_info_returns_204(
        test_collaboration,
        mock_box_session):
    # pylint:disable=protected-access
    data = {'role': CollaborationRole.OWNER, 'status': CollaborationStatus.ACCEPTED}
    expected_url = test_collaboration.get_url()
    mock_box_session.put.return_value.ok = True
    is_success = test_collaboration.update_info(data=data)
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(data),
        expect_json_response=False,
        headers=None,
        params=None,
    )
    assert is_success is True


def test_accept_pending_collaboration(test_collaboration, mock_box_session):
    # pylint:disable=protected-access
    new_status = 'accepted'
    expected_url = f'{API.BASE_API_URL}/collaborations/{test_collaboration.object_id}'
    mock_collab_response = {
        'type': 'collaboration',
        'id': '1234',
        'status': 'accepted',
    }
    mock_box_session.put.return_value.json.return_value = mock_collab_response
    response = test_collaboration.accept()
    update_body = {
        'status': 'accepted'
    }
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(update_body),
        headers=None,
        params=None,
    )
    assert isinstance(response, test_collaboration.__class__)
    assert response.status == new_status


def test_reject_pending_collaboration(test_collaboration, mock_box_session):
    # pylint:disable=protected-access
    new_status = 'rejected'
    expected_url = f'{API.BASE_API_URL}/collaborations/{test_collaboration.object_id}'
    mock_collab_response = {
        'type': 'collaboration',
        'id': '1234',
        'status': 'rejected',
    }
    mock_box_session.put.return_value.json.return_value = mock_collab_response
    response = test_collaboration.reject()
    update_body = {
        'status': 'rejected'
    }
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(update_body),
        headers=None,
        params=None,
    )
    assert isinstance(response, test_collaboration.__class__)
    assert response.status == new_status



from boxsdk.pagination.box_object_collection import BoxObjectCollection

from test.integration_new import CLIENT


def test_get_sign_templates():
    sign_templates = CLIENT.get_sign_templates()
    assert isinstance(sign_templates, BoxObjectCollection)



from datetime import datetime

import pytest
import hashlib
import os

from pytest_lazyfixture import lazy_fixture

from boxsdk.exception import BoxAPIException
from boxsdk.object.collaboration import CollaborationRole
from test.integration_new.context_managers.box_metadata_template import BoxTestMetadataTemplate
from test.integration_new.context_managers.box_test_group import BoxTestGroup
from test.integration_new.context_managers.box_test_web_link import BoxTestWebLink
from test.integration_new import CLIENT
from test.integration_new import util
from test.integration_new.context_managers.box_test_file import BoxTestFile
from test.integration_new.context_managers.box_test_folder import BoxTestFolder

FOLDER_TESTS_DIRECTORY_NAME = 'folder-integration-tests'


@pytest.fixture(scope="module", autouse=True)
def parent_folder():
    with BoxTestFolder(name=f'{FOLDER_TESTS_DIRECTORY_NAME} {datetime.now()}') as folder:
        yield folder


@pytest.fixture(scope="module")
def group():
    with BoxTestGroup() as group:
        yield group


def test_preflight_check(parent_folder):
    file_size = 1213
    accelerator_url = parent_folder.preflight_check(file_size, util.random_name())

    assert accelerator_url


def test_manual_chunked_upload(parent_folder, large_file, large_file_name):
    total_size = os.stat(large_file.path).st_size
    sha1 = hashlib.sha1()
    with open(large_file.path, 'rb') as content_stream:
        upload_session = parent_folder.create_upload_session(file_size=total_size, file_name=large_file_name)
        part_array = []
        for part_num in range(upload_session.total_parts):

            copied_length = 0
            chunk = b''
            while copied_length < upload_session.part_size:
                bytes_read = content_stream.read(upload_session.part_size - copied_length)
                if bytes_read is None:
                    continue
                if len(bytes_read) == 0:
                    break
                chunk += bytes_read
                copied_length += len(bytes_read)

            uploaded_part = upload_session.upload_part_bytes(chunk, part_num * upload_session.part_size, total_size)
            part_array.append(uploaded_part)
            sha1.update(chunk)
        content_sha1 = sha1.digest()
        uploaded_file = upload_session.commit(content_sha1=content_sha1, parts=part_array)

        try:
            assert uploaded_file.id
            assert uploaded_file.name == large_file_name
            assert uploaded_file.parent == parent_folder
            assert uploaded_file.size == total_size
        finally:
            util.permanently_delete(uploaded_file)


def test_auto_chunked_upload(parent_folder, large_file, large_file_name):
    total_size = os.stat(large_file.path).st_size
    chunked_uploader = parent_folder.get_chunked_uploader(large_file.path)

    uploaded_file = chunked_uploader.start()

    try:
        assert uploaded_file.id
        assert uploaded_file.name == large_file_name
        assert uploaded_file.parent == parent_folder
        assert uploaded_file.size == total_size
    finally:
        util.permanently_delete(uploaded_file)


def test_get_items(parent_folder, small_file_path):
    with BoxTestFolder(parent_folder=parent_folder) as subfolder,\
            BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as file,\
            BoxTestWebLink(parent_folder=parent_folder, url='https://box.com') as web_link:

        assert set(parent_folder.get_items()) == {subfolder, file, web_link}


def test_upload_stream_to_folder(parent_folder, small_file_name, small_file_path):
    with open(small_file_path, 'rb') as stream_to_be_uploaded:
        uploaded_file = parent_folder.upload_stream(file_stream=stream_to_be_uploaded, file_name=small_file_name)

        try:
            assert uploaded_file.id
            assert uploaded_file.parent == parent_folder
        finally:
            util.permanently_delete(uploaded_file)


def test_upload_small_file_to_folder(parent_folder, small_file_name, small_file_path):
    uploaded_file = parent_folder.upload(file_path=small_file_path, file_name=small_file_name)
    try:
        assert uploaded_file.id
        assert uploaded_file.parent == parent_folder
    finally:
        util.permanently_delete(uploaded_file)


def test_create_subfolder(parent_folder):
    created_subfolder = parent_folder.create_subfolder(name=util.random_name())
    try:
        assert created_subfolder.id
        assert created_subfolder.parent == parent_folder
    finally:
        util.permanently_delete(created_subfolder)


def test_get_shared_link(parent_folder, other_user, other_client):
    with BoxTestFolder(parent_folder=parent_folder) as folder:
        folder.collaborate(accessible_by=other_user, role='editor')

        shared_link = other_client.folder(folder.object_id).get_shared_link(allow_preview=True, allow_download=True)

        result_permissions = folder.get().shared_link['permissions']
        assert result_permissions == {'can_preview': True, 'can_download': True, 'can_edit': False}
        assert other_client.get_shared_item(shared_link).id == folder.id


@pytest.mark.parametrize(
    'collaborator', [
        lazy_fixture('user'),
        lazy_fixture('group'),
    ]
)
def test_add_collaborator(parent_folder, collaborator):
    with BoxTestFolder(parent_folder=parent_folder) as folder:
        folder.add_collaborator(collaborator=collaborator, role=CollaborationRole.EDITOR)
        assert list(folder.get_collaborations())[0].accessible_by == collaborator


def test_add_collaborator_using_email(parent_folder, user):
    user_email = user.login
    with BoxTestFolder(parent_folder=parent_folder) as folder:
        folder.add_collaborator(collaborator=user_email, role=CollaborationRole.VIEWER)
        assert list(folder.get_collaborations())[0].accessible_by == user


def test_invite_collaboratur_using_when_nonexistent_user_email_provided(parent_folder, user):
    nonexistent_user_email = 'non-existant-user-email@box.com'
    with BoxTestFolder(parent_folder=parent_folder) as folder:
        folder.add_collaborator(collaborator=nonexistent_user_email, role=CollaborationRole.VIEWER)
        assert list(folder.get_collaborations())[0].invite_email == nonexistent_user_email


def test_create_web_link(parent_folder):
    created_web_link = parent_folder.create_web_link(target_url="https://box.com")
    try:
        assert created_web_link.id
        assert created_web_link.parent == parent_folder
    finally:
        util.permanently_delete(created_web_link)


def test_delete_folder(parent_folder):
    with BoxTestFolder(parent_folder=parent_folder) as folder:
        created_subfolder = folder.create_subfolder(name=util.random_name())
        try:
            created_subfolder.create_subfolder(name=util.random_name())

            assert list(folder.get_items())
            created_subfolder.delete(recursive=True)
            assert not list(folder.get_items())
        finally:
            CLIENT.trash().permanently_delete_item(created_subfolder)


def test_cascade_and_get_metadata_cascade_policies(parent_folder):
    with BoxTestMetadataTemplate(display_name="test_template") as metadata_template,\
            BoxTestFolder(parent_folder=parent_folder) as folder:
        folder.cascade_metadata(metadata_template)

        policy_applied_to_folder = list(folder.get_metadata_cascade_policies())[0]

        assert policy_applied_to_folder.scope == metadata_template.scope
        assert policy_applied_to_folder.templateKey == metadata_template.template_key
        assert policy_applied_to_folder.parent == folder


def test_create_and_get_lock(parent_folder):
    with BoxTestFolder(parent_folder=parent_folder) as folder:
        lock = folder.create_lock()

        try:
            assert lock.id == list(folder.get_locks())[0].id
            assert lock.folder == folder

            with pytest.raises(BoxAPIException):
                folder.move(parent_folder=CLIENT.root_folder())

            with pytest.raises(BoxAPIException):
                folder.delete()
        finally:
            lock.delete()



from unittest.mock import call
import pytest
from boxsdk.config import API, Client
from boxsdk.object.user import User


@pytest.fixture
def as_user_headers(mock_user_id, access_token):
    return {
        'Authorization': f'Bearer {access_token}',
        'As-User': mock_user_id,
        'User-Agent': Client.USER_AGENT_STRING,
        'X-Box-UA': Client.BOX_UA_STRING,
    }


def test_client_as_user_causes_as_user_header_to_be_added(
        box_client,
        mock_box_network,
        generic_successful_response,
        mock_user_id,
        as_user_headers,
):
    # pylint:disable=redefined-outer-name
    mock_box_network.session.request.side_effect = [generic_successful_response]
    box_client.as_user(User(None, mock_user_id)).folder('0').get()
    assert mock_box_network.session.request.mock_calls == [
        call(
            'GET',
            f'{API.BASE_API_URL}/folders/0',
            headers=as_user_headers,
            params=None,
            log_response_content=True,
        ),
    ]


def test_folder_object_as_user_causes_as_user_header_to_be_added(
        box_client,
        mock_box_network,
        generic_successful_response,
        mock_user_id,
        as_user_headers,
):
    # pylint:disable=redefined-outer-name
    mock_box_network.session.request.side_effect = [
        generic_successful_response,
    ]
    box_client.folder('0').as_user(User(None, mock_user_id)).get()
    assert mock_box_network.session.request.mock_calls == [
        call(
            'GET',
            f'{API.BASE_API_URL}/folders/0',
            headers=as_user_headers,
            params=None,
            log_response_content=True,
        ),
    ]


def test_group_membership_object_as_user_causes_as_user_header_to_be_added(
        box_client,
        mock_box_network,
        generic_successful_response,
        mock_user_id,
        as_user_headers,
):
    # pylint:disable=redefined-outer-name
    mock_box_network.session.request.side_effect = [
        generic_successful_response,
    ]
    box_client.group_membership('0').as_user(User(None, mock_user_id)).get()
    assert mock_box_network.session.request.mock_calls == [
        call(
            'GET',
            f'{API.BASE_API_URL}/group_memberships/0',
            headers=as_user_headers,
            params=None,
            log_response_content=True,
        ),
    ]


def test_events_endpoint_as_user_causes_as_user_header_to_be_added(
        box_client,
        mock_box_network,
        generic_successful_response,
        mock_user_id,
        as_user_headers,
):
    # pylint:disable=redefined-outer-name
    mock_box_network.session.request.side_effect = [
        generic_successful_response,
    ]
    stream_position = 1348790499819
    options = {'url': f'{API.BASE_API_URL}/events', 'retry_timeout': 60}
    box_client.events().as_user(User(None, mock_user_id)).long_poll(options, stream_position)
    assert mock_box_network.session.request.mock_calls == [
        call(
            'GET',
            options['url'],
            headers=as_user_headers,
            timeout=options['retry_timeout'],
            params={'stream_position': stream_position},
            log_response_content=True,
        ),
    ]


def test_metadata_endpoint_as_user_causes_as_user_header_to_be_added(
        box_client,
        mock_box_network,
        generic_successful_response,
        mock_user_id,
        as_user_headers,
):
    # pylint:disable=redefined-outer-name
    mock_box_network.session.request.side_effect = [
        generic_successful_response,
    ]
    box_client.file('0').metadata().as_user(User(None, mock_user_id)).get()
    assert mock_box_network.session.request.mock_calls == [
        call(
            'GET',
            f'{API.BASE_API_URL}/files/0/metadata/global/properties',
            headers=as_user_headers,
            log_response_content=True,
        ),
    ]



from datetime import datetime

import pytest

from boxsdk.object.metadata_template import MetadataField, MetadataFieldType
from test.integration_new import CLIENT
from test.integration_new.context_managers.box_test_folder import BoxTestFolder

FILE_TESTS_DIRECTORY_NAME = 'metadata_template-integration-tests'


@pytest.fixture(scope="module", autouse=True)
def parent_folder():
    with BoxTestFolder(name=f'{FILE_TESTS_DIRECTORY_NAME} {datetime.now()}') as folder:
        yield folder


def test_create_metedata_template_with_fields(parent_folder):
    metadata_template_fields = [
        MetadataField(
            field_type=MetadataFieldType.ENUM,
            display_name='State',
            key='state',
            description='Which state in USA',
            options=['CA', 'TX', 'NY'],
            hidden=True
        )
    ]

    metadata_template = CLIENT.create_metadata_template(
        display_name="template_name", fields=metadata_template_fields, hidden=False
    )

    try:
        assert metadata_template.displayName == 'template_name'
        assert not metadata_template.hidden

        metadata_template_field = metadata_template.fields[0]
        assert metadata_template_field['type'] == 'enum'
        assert metadata_template_field['displayName'] == 'State'
        assert metadata_template_field['key'] == 'state'
        assert metadata_template_field['description'] == 'Which state in USA'
        assert len(metadata_template_field['options']) == 3
        assert metadata_template_field['hidden']

    finally:
        metadata_template.delete()



import base64
import hashlib
import json
import os
from http import HTTPStatus
from typing import Any, Optional, TYPE_CHECKING, Iterable, IO

from boxsdk import BoxAPIException
from boxsdk.util.api_call_decorator import api_call
from boxsdk.util.chunked_uploader import ChunkedUploader
from .base_object import BaseObject
from ..pagination.limit_offset_based_dict_collection import LimitOffsetBasedDictCollection

if TYPE_CHECKING:
    from boxsdk.pagination.box_object_collection import BoxObjectCollection
    from boxsdk.object.file import File


class UploadSession(BaseObject):
    _item_type = 'upload_session'
    _parent_item_type = 'file'

    def get_url(self, *args: Any) -> str:
        """
        Base class override. Endpoint is a little different - it's /files/upload_sessions.
        """
        return self._session.get_url(
            f'{self._parent_item_type}s/{self._item_type}s',
            self._object_id,
            *args
        ).replace(self.session.api_config.BASE_API_URL, self.session.api_config.UPLOAD_URL)

    @api_call
    def get_parts(self, limit: Optional[int] = None, offset: Optional[int] = None) -> 'BoxObjectCollection':
        """
        Get a list of parts uploaded so far.

        :param limit:
            The maximum number of items to return per page. If not specified, then will use the server-side default.
        :param offset:
            The index at which to start returning items.
        :returns:
            Returns a :class:`BoxObjectCollection` object containing the uploaded parts.
        """
        return LimitOffsetBasedDictCollection(
            session=self.session,
            url=self.get_url('parts'),
            limit=limit,
            offset=offset,
            fields=None,
            return_full_pages=False,
        )

    @api_call
    def upload_part_bytes(
            self,
            part_bytes: bytes,
            offset: int,
            total_size: int,
            part_content_sha1: Optional[bytes] = None
    ) -> dict:
        """
        Upload a part of a file.

        :param part_bytes:
            Part bytes
        :param offset:
            Offset, in number of bytes, of the part compared to the beginning of the file. This number should be a
            multiple of the part size.
        :param total_size:
            The size of the file that this part belongs to.
        :param part_content_sha1:
            SHA-1 hash of the part's content. If not specified, this will be calculated.
        :returns:
            The uploaded part record.
        """

        if part_content_sha1 is None:
            sha1 = hashlib.sha1()
            sha1.update(part_bytes)
            part_content_sha1 = sha1.digest()

        range_end = min(offset + self.part_size - 1, total_size - 1)  # pylint:disable=no-member
        headers = {
            'Content-Type': 'application/octet-stream',
            'Digest': f'SHA={base64.b64encode(part_content_sha1).decode("utf-8")}',
            'Content-Range': f'bytes {offset}-{range_end}/{total_size}',
        }
        response = self._session.put(
            self.get_url(),
            headers=headers,
            data=part_bytes,
        )
        return response.json()['part']

    @api_call
    def commit(
            self,
            content_sha1: bytes,
            parts: Iterable[Optional[dict]] = None,
            file_attributes: dict = None,
            etag: Optional[str] = None
    ) -> Optional['File']:
        """
        Commit a multiput upload.

        :param content_sha1:
            SHA-1 hash of the file contents that was uploaded.
        :param parts:
            List of parts that were uploaded.
        :param file_attributes:
            A `dict` of attributes to set on the uploaded file.
        :param etag:
            If specified, instruct the Box API to delete the folder only if the current version's etag matches.
        :returns:
            The newly-uploaded file object or None if commit was not processed
        """
        body = {}
        if file_attributes is not None:
            body['attributes'] = file_attributes
        if parts is not None:
            body['parts'] = parts
        else:
            body['parts'] = list(self.get_parts())
        headers = {
            'Content-Type': 'application/json',
            'Digest': f'SHA={base64.b64encode(content_sha1).decode("utf-8")}',
        }
        if etag is not None:
            headers['If-Match'] = etag

        try:
            response = self._session.post(
                self.get_url('commit'),
                headers=headers,
                data=json.dumps(body),
            )
        except BoxAPIException as box_api_exc:
            if box_api_exc.status == HTTPStatus.ACCEPTED:
                return None
            raise box_api_exc

        entry = response.json()['entries'][0]
        return self.translator.translate(
            session=self._session,
            response_object=entry,
        )

    @api_call
    def abort(self) -> bool:
        """
        Abort an upload session, cancelling the upload and removing any parts that have already been uploaded.

        :returns:
            A boolean indication success of the upload abort.
        """
        return self.delete()

    def get_chunked_uploader_for_stream(self, content_stream: IO[bytes], file_size: int) -> ChunkedUploader:
        """
        Instantiate the chunked upload instance and create upload session.

        :param content_stream:
            File-like object containing the content of the part to be uploaded.
        :param file_size:
            The size of the file that this part belongs to.
        :returns:
            A :class:`ChunkedUploader` object.
        """
        return ChunkedUploader(self, content_stream, file_size)

    def get_chunked_uploader(self, file_path: str) -> ChunkedUploader:
        # pylint: disable=consider-using-with
        """
        Instantiate the chunked upload instance and create upload session with path to file.

        :param file_path:
            The local path to the file you wish to upload.
        :returns:
            A :class:`ChunkedUploader` object.
        """
        total_size = os.stat(file_path).st_size
        content_stream = open(file_path, 'rb')
        return self.get_chunked_uploader_for_stream(
            content_stream=content_stream,
            file_size=total_size,
        )



import io
import logging
from unittest.mock import mock_open, patch, Mock

import pytest

import boxsdk.util.log


_MOCK_FILEPATH = '/home/user/boxsdk.log'
_MOCK_LOG_NAME = 'boxsdk'


@pytest.fixture(params=[io.StringIO(), _MOCK_FILEPATH])
def stream_or_file(request):
    return request.param


@pytest.fixture(params=[False, True])
def debug(request):
    return request.param


@pytest.fixture
def expected_log_level(debug):
    return logging.DEBUG if debug else logging.INFO


@pytest.fixture(params=[_MOCK_LOG_NAME, None])
def name(request):
    return request.param


@pytest.fixture
def mock_logger():
    return Mock(logging.Logger)


def test_setup_logging(stream_or_file, debug, expected_log_level, name, mock_logger):
    mock_file_open = mock_open()

    with patch('logging.getLogger') as get_logger:
        with patch('logging.open', mock_file_open, create=True):
            get_logger.return_value = mock_logger
            boxsdk.util.log.Logging().setup_logging(stream_or_file, debug=debug, name=name)
            get_logger.assert_called_once_with(name)

    assert mock_logger.addHandler.call_count == 1
    assert isinstance(mock_logger.addHandler.call_args[0][0], logging.Handler)
    mock_logger.setLevel.assert_called_once_with(expected_log_level)

    if isinstance(stream_or_file, str):
        assert mock_file_open.call_count == 1
        assert mock_file_open.call_args[0][:2] == (stream_or_file, 'a')   # Python 3 passes additional args.


def test_setup_logging_is_reentrant(mock_logger):
    mock_file_open = mock_open()

    with patch('logging.getLogger') as get_logger:
        with patch('logging.open', mock_file_open, create=True):
            get_logger.return_value = mock_logger
            logging_instance = boxsdk.util.log.Logging()
            logging_instance.setup_logging(None)
            get_logger.assert_called_once_with(None)
            get_logger.return_value = Mock()
            logging_instance.setup_logging(None)

    assert mock_logger.addHandler.call_count == 1
    assert isinstance(mock_logger.addHandler.call_args[0][0], logging.Handler)
    mock_logger.setLevel.assert_called_once()

    if isinstance(stream_or_file, str):
        assert mock_file_open.call_count == 1


@pytest.mark.parametrize(
    'unsanitized_dict, expected_result',
    [
        # Test for when no sanitization is required
        (
            {'name': 'foo'},
            {'name': 'foo'},
        ),
        # Test for basic string sanitization
        (
            {'access_token': 'askdjfhadsrwedr'},
            {'access_token': '---wedr'},
        ),
        # Test for short string sanitization
        (
            {'refresh_token': 'abc'},
            {'refresh_token': '---abc'},
        ),
        # Test for recursive sanitization
        (
            {'stuff': {'shared_link': 'https://example.com/asdfghjkl'}},
            {'stuff': {'shared_link': '---hjkl'}},
        ),
        # Test for None type
        (
            {'download_url': None},
            {'download_url': None},
        ),
        # Test for proxy http
        (
            {'http': 'http://username:password@localhost:8080'},
            {'http': 'http://---:---@localhost:8080'},
        ),
        # Test for proxy https
        (
            {'https': 'http://username:password@localhost:8080'},
            {'https': 'http://---:---@localhost:8080'},
        ),
    ]
)
def test_sanitize_dictionary_correctly_sanitizes_params(mock_logger, unsanitized_dict, expected_result):
    mock_file_open = mock_open()

    with patch('logging.getLogger') as get_logger:
        with patch('logging.open', mock_file_open, create=True):
            get_logger.return_value = mock_logger
            actual_result = boxsdk.util.log.Logging().sanitize_dictionary(unsanitized_dict)
            assert actual_result == expected_result



import pytest

from boxsdk.config import API
from boxsdk.object.sign_template import SignTemplate


@pytest.fixture(scope='module')
def mock_sign_template_response():
    mock_sign_template = {
        "id": "93153068-5420-467b-b8ef-8e54bfb7be42",
        "type": "sign-template",
        "name": "important-file.pdf",
        "email_message": "Please sign this document.\n\nKind regards",
        "email_subject": "Box User (boxuser@box.com) has requested your signature on a document",
        "parent_folder": {
            "id": "123456789",
            "etag": "0",
            "type": "folder",
            "sequence_id": "0",
            "name": "My Sign Requests"
        },
        "auto_expire_days": "null",
        "source_files": [
            {
                "id": "123456",
                "etag": "0",
                "type": "file",
                "sequence_id": "0",
                "sha1": "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
                "file_version": {
                    "id": "123456",
                    "type": "file_version",
                    "sha1": "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                }
            }
        ],
        "are_email_settings_locked": "false",
        "are_fields_locked": "false",
        "are_files_locked": "false",
        "are_options_locked": "false",
        "are_recipients_locked": "false",
        "signers": [
            {
                "email": "",
                "label": "",
                "public_id": "AAQXQXJZ4",
                "role": "final_copy_reader",
                "is_in_person": "false",
                "order": 1,
                "inputs": []
            },
            {
                "email": "",
                "label": "",
                "public_id": "13XQXJZ4",
                "role": "signer",
                "is_in_person": "false",
                "order": 1,
                "inputs": [
                    {
                        "document_tag_id": None,
                        "id": "0260f921-3b52-477f-ae74-0b0b0b0b0b0b",
                        "type": "signature",
                        "text_value": None,
                        "is_required": True,
                        "coordinates": {
                            "x": 0.27038464059712,
                            "y": 0.10051756244533624
                        },
                        "dimensions": {
                            "width": 0.23570031566618235,
                            "height": 0.04781003891921971
                        },
                        "date_value": None,
                        "page_index": 0,
                        "checkbox_value": None,
                        "document_id": "2fdf9003-d798-40ee-be7f-0b0b0b0b0b0b",
                        "content_type": "signature",
                        "dropdown_choices": None,
                        "group_id": None,
                        "label": None
                    }
                ]
            }
        ],
        "ready_sign_link": None,
        "custom_branding": None,
        "days_valid": 0,
        "additional_info": {
            "non_editable": [],
            "required": {
                "signers": [
                    [
                        "email"
                    ],
                    [
                        "email"
                    ]
                ]
            }
        }
    }
    return mock_sign_template


def test_get_sign_template(test_sign_template, mock_box_session, mock_sign_template_response):
    expected_url = f'{API.BASE_API_URL}/sign_templates/{test_sign_template.object_id}'
    mock_box_session.get.return_value.json.return_value = mock_sign_template_response

    sign_template = test_sign_template.get()

    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)

    assert isinstance(sign_template, SignTemplate)

    assert sign_template.id == '93153068-5420-467b-b8ef-8e54bfb7be42'
    assert sign_template.name == 'important-file.pdf'
    assert sign_template.email_message == 'Please sign this document.\n\nKind regards'



from abc import ABC, abstractmethod
from typing import Any, Callable, Union


class Network(ABC):
    """
    Abstract base class specifying the interface of the network layer.
    """

    @abstractmethod
    def request(self, method: str, url: str, access_token: str, **kwargs: Any) -> 'NetworkResponse':
        """
        Make a network request to the given url with the given method.

        :param method:
            The HTTP verb that should be used to make the request.
        :param url:
            The URL for the request.
        :param access_token:
            The OAuth2 access token used to authorize the request.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def retry_after(self, delay: float, request_method: Callable, *args: Any, **kwargs: Any) -> 'NetworkResponse':
        """
        Make a network request after a given delay.

        :param delay:
            How long until the request should be executed.
        :param request_method:
            A callable that will execute the request.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    def network_response_constructor(self) -> Union[type, Callable]:
        """The constructor to use for creating NetworkResponse instances.

        This is not implemented by default, and is not a required part of the
        interface.

        It is recommended that implementations of `request()` call this to
        construct their responses, rather than hard-coding the construction.
        That way, subclasses of the implementation can easily extend the
        construction of :class:`NetworkResponse` instances, by overriding this
        property, instead of needing to override `request()`.

        :return:
            A callable that returns an instance of :class:`NetworkResponse`.
            Most commonly, this will be a subclass of :class:`NetworkResponse`.
        """
        return NetworkResponse


class NetworkResponse(ABC):
    """Abstract base class specifying the interface for a network response."""

    @abstractmethod
    def json(self) -> Union[dict, list, str, int, float]:
        """
        Return the parsed JSON response.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def content(self) -> Any:
        """
        Return the content of the response body.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def status_code(self) -> int:
        """
        Return the HTTP status code of the response.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def ok(self) -> bool:
        """
        Return whether or not the request was successful.
        """
        # pylint:disable=invalid-name
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def headers(self) -> dict:
        """
        Return the response headers.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def response_as_stream(self) -> Any:
        """
        Return a stream containing the raw network response.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def access_token_used(self) -> str:
        """
        Return the access token used to make the request.
        """
        raise NotImplementedError  # pragma: no cover



import json
from boxsdk.config import API
from boxsdk.object.storage_policy import StoragePolicy
from boxsdk.object.storage_policy_assignment import StoragePolicyAssignment


def test_get(test_storage_policy, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/storage_policies/{test_storage_policy.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': test_storage_policy.object_type,
        'id': test_storage_policy.object_id,
    }
    storage_policy = test_storage_policy.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(storage_policy, StoragePolicy)


def test_assign_with_same_assignment(test_storage_policy, test_storage_policy_assignment, mock_user, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/storage_policy_assignments'
    additional_params = {
        'resolved_for_type': mock_user.object_type,
        'resolved_for_id': mock_user.object_id,
    }
    mock_box_session.get.return_value.json.return_value = {
        'next_marker': None,
        'limit': 1,
        'entries': [
            {
                'type': test_storage_policy_assignment.object_type,
                'id': test_storage_policy_assignment.object_id,
                'storage_policy': {
                    'type': test_storage_policy.object_type,
                    'id': test_storage_policy.object_id,
                },
            },
        ],
    }
    assignment = test_storage_policy.assign(mock_user)
    mock_box_session.get.assert_called_once_with(expected_url, params=additional_params)
    assert isinstance(assignment, StoragePolicyAssignment)
    assert assignment.type == test_storage_policy_assignment.object_type
    assert assignment.id == test_storage_policy_assignment.object_id
    assert assignment.storage_policy['type'] == test_storage_policy.object_type
    assert assignment.storage_policy['id'] == test_storage_policy.object_id


def test_assign_with_assigned_enterprise(test_storage_policy, test_storage_policy_assignment, mock_user, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/storage_policy_assignments'
    additional_params = {
        'resolved_for_type': mock_user.object_type,
        'resolved_for_id': mock_user.object_id,
    }
    expected_data = {
        'storage_policy': {
            'type': test_storage_policy.object_type,
            'id': test_storage_policy.object_id,
        },
        'assigned_to': {
            'type': mock_user.object_type,
            'id': mock_user.object_id,
        },
    }
    mock_box_session.get.return_value.json.return_value = {
        'next_marker': None,
        'limit': 1,
        'entries': [
            {
                'type': test_storage_policy_assignment.object_type,
                'id': '11111',
                'storage_policy': {
                    'type': test_storage_policy.object_type,
                    'id': '22222',
                },
                'assigned_to': {
                    'type': 'enterprise',
                    'id': '12345',
                },
            },
        ],
    }
    mock_box_session.post.return_value.json.return_value = {
        'type': test_storage_policy_assignment.object_type,
        'id': test_storage_policy_assignment.object_id,
        'storage_policy': {
            'type': test_storage_policy.object_type,
            'id': test_storage_policy.object_id
        }
    }
    assignment = test_storage_policy.assign(mock_user)
    mock_box_session.get.assert_called_once_with(expected_url, params=additional_params)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    assert isinstance(assignment, StoragePolicyAssignment)
    assert assignment.type == test_storage_policy_assignment.object_type
    assert assignment.id == test_storage_policy_assignment.object_id
    assert assignment.storage_policy['type'] == test_storage_policy.object_type
    assert assignment.storage_policy['id'] == test_storage_policy.object_id


def test_assign_with_update(test_storage_policy, test_storage_policy_assignment, mock_user, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/storage_policy_assignments'
    expected_put_url = f'{API.BASE_API_URL}/storage_policy_assignments/11111'
    additional_params = {
        'resolved_for_type': mock_user.object_type,
        'resolved_for_id': mock_user.object_id,
    }
    expected_data = {
        'storage_policy': {
            'type': test_storage_policy.object_type,
            'id': '42',
        },
    }
    mock_box_session.get.return_value.json.return_value = {
        'next_marker': None,
        'limit': 1,
        'entries': [
            {
                'type': test_storage_policy_assignment.object_type,
                'id': '11111',
                'storage_policy': {
                    'type': test_storage_policy.object_type,
                    'id': '22222',
                },
                'assigned_to': {
                    'type': 'user',
                    'id': '12345',
                },
            },
        ],
    }
    mock_box_session.put.return_value.json.return_value = {
        'type': test_storage_policy_assignment.object_type,
        'id': test_storage_policy_assignment.object_id,
        'storage_policy': {
            'type': test_storage_policy.object_type,
            'id': '42'
        }
    }
    assignment = test_storage_policy.assign(mock_user)
    mock_box_session.get.assert_called_once_with(expected_url, params=additional_params)
    mock_box_session.put.assert_called_once_with(expected_put_url, data=json.dumps(expected_data), headers=None, params=None)
    assert isinstance(assignment, StoragePolicyAssignment)
    assert assignment.type == test_storage_policy_assignment.object_type
    assert assignment.id == test_storage_policy_assignment.object_id
    assert assignment.storage_policy['type'] == test_storage_policy.object_type
    assert assignment.storage_policy['id'] == '42'


def test_create_assignment(test_storage_policy, mock_user, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/storage_policy_assignments'
    expected_data = {
        'storage_policy': {
            'type': 'storage_policy',
            'id': '42',
        },
        'assigned_to': {
            'type': 'user',
            'id': 'fake-user-100',
        },
    }
    mock_assignment = {
        'type': 'storage_policy_assignment',
        'id': '1234',
        'storage_policy': {
            'type': test_storage_policy.object_type,
            'id': test_storage_policy.object_id,
        },
    }
    mock_box_session.post.return_value.json.return_value = mock_assignment
    assignment = test_storage_policy.create_assignment(mock_user)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    assert assignment.id == mock_assignment['id']
    assert assignment.storage_policy['id'] == mock_assignment['storage_policy']['id']



import random
import math

from functools import partial
from logging import getLogger
from numbers import Number
from typing import TYPE_CHECKING, Optional, Any, Type, Callable, Set

from requests.exceptions import RequestException
from boxsdk.exception import BoxException
from .box_request import BoxRequest as _BoxRequest
from .box_response import BoxResponse as _BoxResponse
from ..config import API, Client, Proxy
from ..exception import BoxAPIException
from ..network.default_network import DefaultNetwork
from ..util.json import is_json_response
from ..util.multipart_stream import MultipartStream
from ..util.shared_link import get_shared_link_header
from ..util.translator import Translator

if TYPE_CHECKING:
    from boxsdk.network.network_interface import Network
    from boxsdk.object.user import User
    from boxsdk import NetworkResponse, OAuth2


class Session:

    _retry_randomization_factor = 0.5
    _retry_base_interval = 1
    _JWT_GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:jwt-bearer'
    _CCG_GRANT_TYPE = 'client_credentials'

    """
    Box API session. Provides automatic retry of failed requests.
    """
    def __init__(
            self,
            network_layer: 'Network' = None,
            default_headers: Optional['dict'] = None,
            translator: Translator = None,
            default_network_request_kwargs: Optional['dict'] = None,
            api_config: API = None,
            client_config: Client = None,
            proxy_config: Optional[Proxy] = None,
    ):
        """
        :param network_layer:
            Network implementation used by the session to make requests.
        :param default_headers:
            A dictionary containing default values to be used as headers when this session makes an API request.
        :param translator:
            (optional) The translator to use for translating Box API JSON
            responses into :class:`BaseAPIJSONObject` smart objects.
            Defaults to a new :class:`Translator` that inherits the
            registrations of the default translator.
        :param default_network_request_kwargs:
            A dictionary containing default values to be passed to the network layer
            when this session makes an API request.
        :param api_config:
            Object containing URLs for the Box API.
        :param client_config:
            Object containing client information, including user agent string.
        :param proxy_config:
            Object containing proxy information.
        """
        if translator is None:
            translator = Translator(extend_default_translator=True, new_child=True)
        self._api_config = api_config or API()
        self._client_config = client_config or Client()
        self._proxy_config = proxy_config or Proxy()
        super().__init__()
        self._network_layer = network_layer or DefaultNetwork()
        self._default_headers = {
            'User-Agent': self._client_config.USER_AGENT_STRING,
            'X-Box-UA': self._client_config.BOX_UA_STRING,
        }
        self._translator = translator
        self._default_network_request_kwargs = {}
        if default_headers:
            self._default_headers.update(default_headers)
        if default_network_request_kwargs:
            self._default_network_request_kwargs.update(default_network_request_kwargs)
        self._logger = getLogger(__name__)

    def get(self, url: str, **kwargs: Any) -> '_BoxResponse':
        """Make a GET request to the Box API.

        :param url:
            The URL for the request.
        """
        return self.request('GET', url, **kwargs)

    def post(self, url: str, **kwargs: Any) -> '_BoxResponse':
        """Make a POST request to the Box API.

        :param url:
            The URL for the request.
        """
        return self.request('POST', url, **kwargs)

    def put(self, url: str, **kwargs: Any) -> '_BoxResponse':
        """Make a PUT request to the Box API.

        :param url:
            The URL for the request.
        """
        return self.request('PUT', url, **kwargs)

    def delete(self, url: str, **kwargs: Any) -> '_BoxResponse':
        """Make a DELETE request to the Box API.

        :param url:
            The URL for the request.
        """
        if 'expect_json_response' not in kwargs:
            kwargs['expect_json_response'] = False
        if 'skip_retry_codes' not in kwargs:
            kwargs['skip_retry_codes'] = {202}
        return self.request('DELETE', url, **kwargs)

    def options(self, url: str, **kwargs: Any) -> '_BoxResponse':
        """Make an OPTIONS request to the Box API.

        :param url:
            The URL for the request.
        """
        return self.request('OPTIONS', url, **kwargs)

    def request(self, method: str, url: str, **kwargs: Any) -> '_BoxResponse':
        """Make a request to the Box API.

        :param method:
            The HTTP verb for the request.
        :param url:
            The URL for the request.
        """
        response = self._prepare_and_send_request(method, url, **kwargs)
        return self.box_response_constructor(response)

    @property
    def box_request_constructor(self) -> Type[_BoxRequest]:
        """Get the constructor for the container class representing an API request"""
        return _BoxRequest

    @property
    def box_response_constructor(self) -> Type[_BoxResponse]:
        """Get the constructor for the container class representing an API response"""
        return _BoxResponse

    @property
    def translator(self) -> Translator:
        """
        The translator used for translating Box API JSON responses into `BaseAPIJSONObject` smart objects.
        """
        return self._translator

    @property
    def api_config(self) -> API:
        return self._api_config

    @property
    def client_config(self) -> Client:
        return self._client_config

    @property
    def proxy_config(self) -> Proxy:
        return self._proxy_config

    def get_url(self, endpoint: str, *args: Any) -> str:
        """
        Return the URL for the given Box API endpoint.

        :param endpoint:
            The name of the endpoint.
        :param args:
            Additional parts of the endpoint URL.
        """
        # pylint:disable=no-self-use
        url = [f'{self._api_config.BASE_API_URL}/{endpoint}']
        url.extend([f'/{x}' for x in args])
        return ''.join(url)

    def get_constructor_kwargs(self) -> dict:
        return dict(
            network_layer=self._network_layer,
            translator=self._translator,
            default_network_request_kwargs=self._default_network_request_kwargs.copy(),
            api_config=self._api_config,
            client_config=self._client_config,
            proxy_config=self._proxy_config,
            default_headers=self._default_headers.copy(),
        )

    def as_user(self, user: 'User') -> 'Session':
        """
        Returns a new session object with default headers set up to make requests as the specified user.

        :param user:
            The user to impersonate when making API requests.
        """
        kwargs = self.get_constructor_kwargs()
        kwargs['default_headers']['As-User'] = user.object_id
        return self.__class__(**kwargs)

    def with_shared_link(self, shared_link: str, shared_link_password: str = None) -> 'Session':
        """
        Returns a new session object with default headers set up to make requests using the shared link for auth.

        :param shared_link:
            The shared link.
        :param shared_link_password:
            The password for the shared link.
        """
        kwargs = self.get_constructor_kwargs()
        kwargs['default_headers'].update(get_shared_link_header(shared_link, shared_link_password))
        return self.__class__(**kwargs)

    def with_default_network_request_kwargs(self, extra_network_parameters: dict) -> 'Session':
        kwargs = self.get_constructor_kwargs()
        kwargs['default_network_request_kwargs'].update(extra_network_parameters)
        return self.__class__(**kwargs)

    # We updated our retry strategy to use exponential backoff instead of the header returned from the API response.
    # This is something we can remove in latter major bumps.
    # pylint: disable=unused-argument
    def get_retry_after_time(self, attempt_number: int, retry_after_header: Optional[str]) -> Number:
        """
        Get the amount of time to wait before retrying the API request, using the attempt number that failed to
        calculate the retry time for the next retry attempt.

        If the Retry-After header is supplied, use it; otherwise, use exponential backoff
        For 202 Accepted (thumbnail or file not ready) and 429 (too many requests), retry later, after a delay
        specified by the Retry-After header.
        For 5xx Server Error, retry later, after a delay; use exponential backoff to determine the delay.

        :param attempt_number:          How many attempts at this request have already been tried.
        :param retry_after_header:      Value of the 'Retry-After` response header.
        :return:                        Number of seconds to wait before retrying.
        """
        if retry_after_header is not None:
            try:
                return int(retry_after_header)
            except (ValueError, TypeError):
                pass
        min_randomization = 1 - self._retry_randomization_factor
        max_randomization = 1 + self._retry_randomization_factor
        randomization = (random.uniform(0, 1) * (max_randomization - min_randomization)) + min_randomization
        exponential = math.pow(2, attempt_number)
        return exponential * self._retry_base_interval * randomization

    @staticmethod
    def _raise_on_unsuccessful_request(network_response: 'NetworkResponse', request: '_BoxRequest', raised_exception: Exception) -> None:
        """
        Raise an exception if the request was unsuccessful.

        :param network_response:
            The network response which is being tested for success.
        :param request:
            The API request that could be unsuccessful.
        """
        if network_response is None:
            raise raised_exception

        if not network_response.ok:
            response_json = {}
            try:
                response_json = network_response.json()
            except ValueError:
                pass
            raise BoxAPIException(
                status=network_response.status_code,
                headers=network_response.headers,
                code=response_json.get('code', None) or response_json.get('error', None),
                message=response_json.get('message', None) or response_json.get('error_description', None),
                request_id=response_json.get('request_id', None),
                url=request.url,
                method=request.method,
                context_info=response_json.get('context_info', None),
                network_response=network_response
            )

        if not Session._is_json_response_if_expected(network_response, request):
            raise BoxAPIException(
                status=network_response.status_code,
                headers=network_response.headers,
                message='Non-json response received, while expecting json response.',
                url=request.url,
                method=request.method,
                network_response=network_response,
            )

    @staticmethod
    def _is_json_response_if_expected(network_response: 'NetworkResponse', request: '_BoxRequest') -> bool:
        """
        Validate that the response is json if the request expects json response.

        :param network_response:
            The network response which is being tested for success.
        :param request:
            The API request that could be unsuccessful.
        """
        return not request.expect_json_response or is_json_response(network_response)

    def _prepare_and_send_request(
            self,
            method: str,
            url: str,
            headers: dict = None,
            auto_session_renewal: bool = True,
            expect_json_response: bool = True,
            **kwargs: Any
    ) -> 'NetworkResponse':
        """
        Prepare a request to be sent to the Box API.

        :param method:
            The HTTP verb to use to make the request.
        :param url:
            The request URL.
        :param headers:
            Headers to include with the request.
        :param auto_session_renewal:
            Whether to automatically renew the session if the request fails due to an expired access token.
        :param expect_json_response:
            Whether the response content should be json.
        """
        files = kwargs.get('files')
        kwargs['file_stream_positions'] = None
        if files:
            kwargs['file_stream_positions'] = {name: file_tuple[1].tell() for name, file_tuple in files.items()}
        attempt_number = 0
        request_headers = self._get_request_headers()
        request_headers.update(headers or {})

        request = self.box_request_constructor(
            url=url,
            method=method,
            headers=request_headers,
            auto_session_renewal=auto_session_renewal,
            expect_json_response=expect_json_response,
        )

        skip_retry_codes = kwargs.pop('skip_retry_codes', set())

        raised_exception = None
        try:
            network_response = self._send_request(request, **kwargs)
            reauthentication_needed = network_response.status_code == 401
        except RequestException as request_exc:
            raised_exception = request_exc
            network_response = None
            if 'EOF occurred in violation of protocol' in str(request_exc):
                reauthentication_needed = True
            elif any(text in str(request_exc) for text in [
                'Connection aborted', 'Connection broken', 'Connection reset'
            ]):
                reauthentication_needed = False
            else:
                raise

        while True:
            retry = self._get_retry_request_callable(
                network_response, attempt_number, request, skip_retry_codes, reauthentication_needed, **kwargs)

            if retry is None or attempt_number >= API.MAX_RETRY_ATTEMPTS:
                if network_response is None:
                    raise raised_exception
                break

            attempt_number += 1
            self._logger.debug('Retrying request')
            network_response = retry(request, **kwargs)

        self._raise_on_unsuccessful_request(network_response, request, raised_exception)

        return network_response

    def _get_retry_request_callable(
            self,
            network_response: Optional['NetworkResponse'],
            attempt_number: int,
            request: '_BoxRequest',
            skip_retry_codes: Set[int],
            session_renewal_needed: bool = False,
            **kwargs: Any
    ) -> Optional[Callable]:
        """
        Get a callable that retries a request for certain types of failure.

        For 202 Accepted (thumbnail or file not ready) and 429 (too many requests), retry later, after a delay
        specified by the Retry-After header.
        For 5xx Server Error, retry later, after a delay; use exponential backoff to determine the delay.

        Otherwise, return None.

        :param network_response:
            The response from the Box API.
        :param attempt_number:
            How many attempts at this request have already been tried. Used for exponential backoff calculations.
        :param request:
            The API request that could require retrying.
        :return:
            Callable that, when called, will retry the request. Takes the same parameters as :meth:`_send_request`.
        """
        # pylint:disable=unused-argument
        # pylint:disable=line-too-long
        if network_response is None or (network_response.ok and request.method == 'GET' and not self._is_json_response_if_expected(network_response, request)):
            return partial(
                self._network_layer.retry_after,
                self.get_retry_after_time(attempt_number, None),
                self._send_request,
            )
        code = network_response.status_code

        if (code in (202, 429) or code >= 500) and code not in skip_retry_codes and not self._is_server_auth_type(kwargs):
            return partial(
                self._network_layer.retry_after,
                self.get_retry_after_time(attempt_number, network_response.headers.get('Retry-After', None)),
                self._send_request,
            )
        return None

    def _is_server_auth_type(self, kwargs: dict) -> bool:
        data = kwargs.get('data', {})
        grant_type = None
        try:
            if 'grant_type' in data:
                grant_type = data['grant_type']
        except TypeError:
            pass
        return grant_type in (self._JWT_GRANT_TYPE, self._CCG_GRANT_TYPE)

    def _get_request_headers(self) -> dict:
        return self._default_headers.copy()

    def _prepare_proxy(self) -> Optional[dict]:
        """
        Prepares basic authenticated and unauthenticated proxies for requests.

        :return:
            A prepared proxy dict to send along with the request. None if incorrect parameters were passed.
        """
        proxy = {}
        if self._proxy_config.URL is None:
            return None
        if self._proxy_config.AUTH and {'user', 'password'} <= set(self._proxy_config.AUTH):
            host = self._proxy_config.URL
            address = host.split('//')[1]
            proxy_string = f'http://{self._proxy_config.AUTH.get("user", None)}:' \
                           f'{self._proxy_config.AUTH.get("password", None)}@{address}'
        elif self._proxy_config.AUTH is None:
            proxy_string = self._proxy_config.URL
        else:
            raise BoxException("The proxy auth dict you provided does not match pattern "
                               "{'user': 'example_user', 'password': 'example_password'}")
        proxy['http'] = proxy_string
        proxy['https'] = proxy['http']

        return proxy

    def _send_request(self, request: '_BoxRequest', **kwargs: Any) -> 'NetworkResponse':
        """
        Make a request to the Box API.

        :param request:
            The API request to send.
        """
        # Reset stream positions to what they were when the request was made so the same data is sent even if this
        # is a retried attempt.
        files, file_stream_positions = kwargs.get('files'), kwargs.pop('file_stream_positions')
        request_kwargs = self._default_network_request_kwargs.copy()
        request_kwargs.update(kwargs)
        proxy_dict = self._prepare_proxy()
        if proxy_dict is not None:
            request_kwargs.update({'proxies': proxy_dict})
        if files and file_stream_positions:
            for name, position in file_stream_positions.items():
                files[name][1].seek(position)
            data = request_kwargs.pop('data', {})
            multipart_stream = MultipartStream(data, files)
            request_kwargs['data'] = multipart_stream
            del request_kwargs['files']
            request.headers['Content-Type'] = multipart_stream.content_type
        request.access_token = request_kwargs.pop('access_token', None)

        # send the request
        network_response = self._network_layer.request(
            request.method,
            request.url,
            access_token=request.access_token,
            headers=request.headers,
            log_response_content=request.expect_json_response,
            **request_kwargs
        )

        return network_response


class AuthorizedSession(Session):
    """
    Box API authorized session. Provides auth, automatic retry of failed requests, and session renewal.
    """

    def __init__(self, oauth: 'OAuth2', **kwargs: Any):
        """
        :param oauth:
            OAuth2 object used by the session to authorize requests.
        :param session:
            The Box API session to wrap for authorization.
        """
        super().__init__(**kwargs)
        self._oauth = oauth

    def get_constructor_kwargs(self) -> dict:
        kwargs = super().get_constructor_kwargs()
        kwargs['oauth'] = self._oauth
        return kwargs

    def _renew_session(self, access_token_used: Optional[str]) -> str:
        """
        Renews the session by refreshing the access token.

        :param access_token_used:
            The access token that's currently being used by the session, that needs to be refreshed.
        """
        new_access_token, _ = self._oauth.refresh(access_token_used)
        return new_access_token

    def _get_retry_request_callable(
            self,
            network_response: Optional['NetworkResponse'],
            attempt_number: int,
            request: '_BoxRequest',
            skip_retry_codes: Set[int],
            session_renewal_needed: bool = False,
            **kwargs: Any
    ) -> Callable:
        """
        Get a callable that retries a request for certain types of failure.

        For 401 Unauthorized responses, renew the session by refreshing the access token; then retry.

        Otherwise, defer to baseclass implementation.

        :param network_response:
            The response from the Box API.
        :param attempt_number:
            How many attempts at this request have already been tried. Used for exponential backoff calculations.
        :param request:
            The API request that could require retrying.
        :return:
            Callable that, when called, will retry the request. Takes the same parameters as :meth:`_send_request`.
        """
        if request.auto_session_renewal and session_renewal_needed:
            self._renew_session(request.access_token)
            request.auto_session_renewal = False
            return self._send_request

        return super()._get_retry_request_callable(
            network_response,
            attempt_number,
            request,
            skip_retry_codes,
            session_renewal_needed,
            **kwargs
        )

    def _send_request(self, request: '_BoxRequest', **kwargs: Any) -> 'NetworkResponse':
        """
        Make a request to the Box API.

        :param request:
            The API request to send.
        """
        # Since there can be session renewal happening in the middle of preparing the request, it's important to be
        # consistent with the access_token being used in the request.
        access_token = self._oauth.access_token
        if request.auto_session_renewal and access_token is None:
            access_token = self._renew_session(None)
            request.auto_session_renewal = False
        authorization_header = {'Authorization': f'Bearer {access_token}'}
        request.headers.update(authorization_header)
        kwargs['access_token'] = access_token
        return super()._send_request(request, **kwargs)



from sys import version_info as py_version

from . import version


class API:
    """Configuration object containing the URLs for the Box API."""
    BASE_API_URL = 'https://api.box.com/2.0'
    UPLOAD_URL = 'https://upload.box.com/api/2.0'
    OAUTH2_API_URL = 'https://api.box.com/oauth2'  # <https://developer.box.com/reference/post-oauth2-token>
    OAUTH2_AUTHORIZE_URL = 'https://account.box.com/api/oauth2/authorize'  # <https://developer.box.com/reference/get-authorize/>
    MAX_RETRY_ATTEMPTS = 5
    CHUNK_UPLOAD_THREADS = 5


class Client:
    """Configuration object containing the user agent string."""
    VERSION = version.__version__
    USER_AGENT_STRING = f'box-python-sdk-{VERSION}'
    BOX_UA_STRING = f'agent=box-python-sdk/{VERSION}; ' \
                    f'env=python/{py_version.major}.{py_version.minor}.{py_version.micro}'


class Proxy:
    URL = None
    AUTH = None



from unittest.mock import Mock
import pytest

from boxsdk.exception import BoxAPIException, BoxOAuthException
from boxsdk.network.default_network import DefaultNetworkResponse


def test_box_api_exception():
    status = 400
    code = 'code'
    message = 'message'
    request_id = '12345'
    headers = {'header': 'value'}
    url = 'https://example.com'
    method = 'GET'
    context_info = {'context': 'value'}
    box_exception = BoxAPIException(
        status,
        code=code,
        message=message,
        request_id=request_id,
        headers=headers,
        url=url,
        method=method,
        context_info=context_info,
    )
    assert box_exception.status == status
    assert box_exception.code == code
    assert box_exception.message == message
    assert box_exception.request_id == request_id
    assert box_exception.headers == headers  # pylint:disable=protected-access
    assert box_exception.url == url
    assert box_exception.method == method
    assert box_exception.context_info == context_info
    assert str(box_exception) == f'''Message: {message}
Status: {status}
Code: {code}
Request ID: {request_id}
Headers: {headers}
URL: {url}
Method: {method}
Context Info: {context_info}'''


@pytest.mark.parametrize('has_network_response', [True, False])
def test_box_oauth_exception(has_network_response):
    status = 400
    message = 'message'
    url = 'https://example.com'
    method = 'GET'
    headers = {'header': 'value'}
    network_response = Mock(DefaultNetworkResponse, headers=headers) if has_network_response else None
    box_exception = BoxOAuthException(
        status,
        message=message,
        url=url,
        method=method,
        network_response=network_response,
    )
    assert str(box_exception) == f'''
Message: {message}
Status: {status}
URL: {url}
Method: {method}
Headers: {headers if has_network_response else 'N/A'}'''
    assert box_exception.network_response is network_response



from typing import Callable, Any, Tuple, Optional

from .oauth2 import OAuth2


class DeveloperTokenAuth(OAuth2):
    ENTER_TOKEN_PROMPT = 'Enter developer token: '

    def __init__(self, get_new_token_callback: Callable[[], str] = None, **kwargs: Any):
        self._get_new_token = get_new_token_callback
        super().__init__(
            client_id=None,
            client_secret=None,
            access_token=self._refresh_developer_token(),
            **kwargs
        )

    def _refresh_developer_token(self) -> str:
        if self._get_new_token is not None:
            return self._get_new_token()

        return input(self.ENTER_TOKEN_PROMPT)

    def _refresh(self, access_token: str) -> Tuple[str, Optional[str]]:
        """
        Base class override.
        Ask for a new developer token.
        """
        self._access_token = self._refresh_developer_token()
        return self._access_token, None

    def revoke(self) -> None:
        """
        Base class override.
        Do nothing; developer tokens can't be revoked without client ID and secret.
        """
